@@ -1,7 +1,5 @@
 /* expression.c: A numeric expression
-   Copyright (c) 2003 Philip Kendall
-
-   $Id: expression.c,v 1.13 2004/06/16 14:05:38 pak21 Exp $
+   Copyright (c) 2003-2016 Philip Kendall
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -10,17 +8,16 @@
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -28,17 +25,21 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 
 #include "debugger_internals.h"
 #include "fuse.h"
+#include "mempool.h"
 #include "ui/ui.h"
+#include "utils.h"
 
 typedef enum expression_type {
 
   DEBUGGER_EXPRESSION_TYPE_INTEGER,
-  DEBUGGER_EXPRESSION_TYPE_REGISTER,
   DEBUGGER_EXPRESSION_TYPE_UNARYOP,
   DEBUGGER_EXPRESSION_TYPE_BINARYOP,
+  DEBUGGER_EXPRESSION_TYPE_SYSVAR,
+  DEBUGGER_EXPRESSION_TYPE_VARIABLE,
 
 } expression_type;
 
@@ -51,10 +52,11 @@ enum precedence_t {
   PRECEDENCE_BITWISE_XOR,
   PRECEDENCE_BITWISE_AND,
   PRECEDENCE_EQUALITY,
-  PRECEDENCE_COMPARISION,
+  PRECEDENCE_COMPARISON,
   PRECEDENCE_ADDITION,
   PRECEDENCE_MULTIPLICATION,
   PRECEDENCE_NEGATE,
+  PRECEDENCE_DEREFERENCE,
 
   PRECEDENCE_ATOMIC,
   /* Highest precedence */
@@ -82,9 +84,10 @@ struct debugger_expression {
 
   union {
     int integer;
-    int reg;
     struct unaryop_type unaryop;
     struct binaryop_type binaryop;
+    char *variable;
+    int system_variable;
   } types;
 
 };
@@ -107,6 +110,9 @@ unaryop_precedence( int operation )
 
   case '!': case '~': case '-': return PRECEDENCE_NEGATE;
 
+  case DEBUGGER_TOKEN_DEREFERENCE:
+    return PRECEDENCE_DEREFERENCE;
+
   default:
     ui_error( UI_ERROR_ERROR, "unknown unary operator %d", operation );
     fuse_abort();
@@ -133,7 +139,7 @@ binaryop_precedence( int operation )
   case    '<': case    '>':
   case DEBUGGER_TOKEN_LESS_THAN_OR_EQUAL_TO:
   case DEBUGGER_TOKEN_GREATER_THAN_OR_EQUAL_TO:
-    return PRECEDENCE_COMPARISION;
+    return PRECEDENCE_COMPARISON;
 
   default:
     ui_error( UI_ERROR_ERROR, "unknown binary operator %d", operation );
@@ -142,15 +148,11 @@ binaryop_precedence( int operation )
 }
 
 debugger_expression*
-debugger_expression_new_number( libspectrum_dword number )
+debugger_expression_new_number( libspectrum_dword number, int pool )
 {
   debugger_expression *exp;
 
-  exp = malloc( sizeof( debugger_expression ) );
-  if( !exp ) {
-    ui_error( UI_ERROR_ERROR, "out of memory at %s:%d", __FILE__, __LINE__ );
-    return NULL;
-  }
+  exp = mempool_new( pool, debugger_expression, 1 );
 
   exp->type = DEBUGGER_EXPRESSION_TYPE_INTEGER;
   exp->precedence = PRECEDENCE_ATOMIC;
@@ -159,35 +161,24 @@ debugger_expression_new_number( libspectrum_dword number )
   return exp;
 }
 
+/* Create a new expression representing a Z80 register
+
+   Now just a pass through to the appropriate system variable.
+   This function should be removed in Fuse 1.4 */
 debugger_expression*
-debugger_expression_new_register( int which )
+debugger_expression_new_register( const char *which, int pool )
 {
-  debugger_expression *exp;
-
-  exp = malloc( sizeof( debugger_expression ) );
-  if( !exp ) {
-    ui_error( UI_ERROR_ERROR, "out of memory at %s:%d", __FILE__, __LINE__ );
-    return NULL;
-  }
-
-  exp->type = DEBUGGER_EXPRESSION_TYPE_REGISTER;
-  exp->precedence = PRECEDENCE_ATOMIC;
-  exp->types.reg = which;
-
-  return exp;
+  return debugger_expression_new_system_variable(
+    debugger_z80_system_variable_type, which, pool );
 }
 
 debugger_expression*
 debugger_expression_new_binaryop( int operation, debugger_expression *operand1,
-				  debugger_expression *operand2 )
+				  debugger_expression *operand2, int pool )
 {
   debugger_expression *exp;
 
-  exp = malloc( sizeof( debugger_expression ) );
-  if( !exp ) {
-    ui_error( UI_ERROR_ERROR, "out of memory at %s:%d", __FILE__, __LINE__ );
-    return NULL;
-  }
+  exp = mempool_new( pool, debugger_expression, 1 );
 
   exp->type = DEBUGGER_EXPRESSION_TYPE_BINARYOP;
   exp->precedence = binaryop_precedence( operation );
@@ -199,17 +190,13 @@ debugger_expression_new_binaryop( int operation, debugger_expression *operand1,
   return exp;
 }
 
-
 debugger_expression*
-debugger_expression_new_unaryop( int operation, debugger_expression *operand )
+debugger_expression_new_unaryop( int operation, debugger_expression *operand,
+				 int pool )
 {
   debugger_expression *exp;
 
-  exp = malloc( sizeof( debugger_expression ) );
-  if( !exp ) {
-    ui_error( UI_ERROR_ERROR, "out of memory at %s:%d", __FILE__, __LINE__ );
-    return NULL;
-  }
+  exp = mempool_new( pool, debugger_expression, 1 );
 
   exp->type = DEBUGGER_EXPRESSION_TYPE_UNARYOP;
   exp->precedence = unaryop_precedence( operation );
@@ -220,6 +207,42 @@ debugger_expression_new_unaryop( int operation, debugger_expression *operand )
   return exp;
 }
 
+debugger_expression*
+debugger_expression_new_system_variable( const char *type, const char *detail,
+                                         int pool )
+{
+  debugger_expression *exp;
+  int system_variable;
+
+  system_variable = debugger_system_variable_find( type, detail );
+  if( system_variable == -1 ) {
+    ui_error( UI_ERROR_WARNING, "System variable %s:%s not known", type,
+              detail );
+    return NULL;
+  }
+
+  exp = mempool_new( pool, debugger_expression, 1 );
+
+  exp->type = DEBUGGER_EXPRESSION_TYPE_SYSVAR;
+  exp->precedence = PRECEDENCE_ATOMIC;
+  exp->types.system_variable = system_variable;
+
+  return exp;
+}
+
+debugger_expression*
+debugger_expression_new_variable( const char *name, int pool )
+{
+  debugger_expression *exp;
+
+  exp = mempool_new( pool, debugger_expression, 1 );
+
+  exp->type = DEBUGGER_EXPRESSION_TYPE_VARIABLE;
+  exp->precedence = PRECEDENCE_ATOMIC;
+  exp->types.variable = mempool_strdup( pool, name );
+
+  return exp;
+}
 
 void
 debugger_expression_delete( debugger_expression *exp )
@@ -227,7 +250,7 @@ debugger_expression_delete( debugger_expression *exp )
   switch( exp->type ) {
     
   case DEBUGGER_EXPRESSION_TYPE_INTEGER:
-  case DEBUGGER_EXPRESSION_TYPE_REGISTER:
+  case DEBUGGER_EXPRESSION_TYPE_SYSVAR:
     break;
 
   case DEBUGGER_EXPRESSION_TYPE_UNARYOP:
@@ -239,9 +262,67 @@ debugger_expression_delete( debugger_expression *exp )
     debugger_expression_delete( exp->types.binaryop.op2 );
     break;
 
+  case DEBUGGER_EXPRESSION_TYPE_VARIABLE:
+    libspectrum_free( exp->types.variable );
+    break;
   }
     
-  free( exp );
+  libspectrum_free( exp );
+}
+
+debugger_expression*
+debugger_expression_copy( debugger_expression *src )
+{
+  debugger_expression *dest;
+
+  dest = libspectrum_new( debugger_expression, 1 );
+  if( !dest ) return NULL;
+
+  dest->type = src->type;
+  dest->precedence = src->precedence;
+
+  switch( dest->type ) {
+
+  case DEBUGGER_EXPRESSION_TYPE_INTEGER:
+    dest->types.integer = src->types.integer;
+    break;
+
+  case DEBUGGER_EXPRESSION_TYPE_UNARYOP:
+    dest->types.unaryop.operation = src->types.unaryop.operation;
+    dest->types.unaryop.op = debugger_expression_copy( src->types.unaryop.op );
+    if( !dest->types.unaryop.op ) {
+      libspectrum_free( dest );
+      return NULL;
+    }
+    break;
+
+  case DEBUGGER_EXPRESSION_TYPE_BINARYOP:
+    dest->types.binaryop.operation = src->types.binaryop.operation;
+    dest->types.binaryop.op1 =
+      debugger_expression_copy( src->types.binaryop.op1 );
+    if( !dest->types.binaryop.op1 ) {
+      libspectrum_free( dest );
+      return NULL;
+    }
+    dest->types.binaryop.op2 =
+      debugger_expression_copy( src->types.binaryop.op2 );
+    if( !dest->types.binaryop.op2 ) {
+      debugger_expression_delete( dest->types.binaryop.op1 );
+      libspectrum_free( dest );
+      return NULL;
+    }
+    break;
+
+  case DEBUGGER_EXPRESSION_TYPE_SYSVAR:
+    dest->types.system_variable = src->types.system_variable;
+    break;
+
+  case DEBUGGER_EXPRESSION_TYPE_VARIABLE:
+    dest->types.variable = utils_safe_strdup( src->types.variable );
+    break;
+  }
+
+  return dest;
 }
 
 libspectrum_dword
@@ -252,15 +333,18 @@ debugger_expression_evaluate( debugger_expression *exp )
   case DEBUGGER_EXPRESSION_TYPE_INTEGER:
     return exp->types.integer;
 
-  case DEBUGGER_EXPRESSION_TYPE_REGISTER:
-    return debugger_register_get( exp->types.reg );
-
   case DEBUGGER_EXPRESSION_TYPE_UNARYOP:
     return evaluate_unaryop( &( exp->types.unaryop ) );
 
   case DEBUGGER_EXPRESSION_TYPE_BINARYOP:
     return evaluate_binaryop( &( exp->types.binaryop ) );
 
+  case DEBUGGER_EXPRESSION_TYPE_SYSVAR:
+    return debugger_system_variable_get( exp->types.system_variable );
+
+  case DEBUGGER_EXPRESSION_TYPE_VARIABLE:
+    return debugger_variable_get( exp->types.variable );
+
   }
 
   ui_error( UI_ERROR_ERROR, "unknown expression type %d", exp->type );
@@ -276,6 +360,9 @@ evaluate_unaryop( struct unaryop_type *unary )
   case '~': return ~debugger_expression_evaluate( unary->op );
   case '-': return -debugger_expression_evaluate( unary->op );
 
+  case DEBUGGER_TOKEN_DEREFERENCE:
+    return readbyte_internal( debugger_expression_evaluate( unary->op ) );
+
   }
 
   ui_error( UI_ERROR_ERROR, "unknown unary operator %d", unary->operation );
@@ -296,8 +383,14 @@ evaluate_binaryop( struct binaryop_type *binary )
   case '*': return debugger_expression_evaluate( binary->op1 ) *
 		   debugger_expression_evaluate( binary->op2 );
 
-  case '/': return debugger_expression_evaluate( binary->op1 ) /
-		   debugger_expression_evaluate( binary->op2 );
+  case '/': {
+      libspectrum_dword op2 = debugger_expression_evaluate( binary->op2 );
+      if( op2 == 0 ) {
+        ui_error( UI_ERROR_ERROR, "divide by 0" );
+        return 0;
+      }
+      return debugger_expression_evaluate( binary->op1 ) / op2;
+    }
 
   case DEBUGGER_TOKEN_EQUAL_TO:
             return debugger_expression_evaluate( binary->op1 ) ==
@@ -358,16 +451,20 @@ debugger_expression_deparse( char *buffer, size_t length,
     }
     return 0;
 
-  case DEBUGGER_EXPRESSION_TYPE_REGISTER:
-    snprintf( buffer, length, "%s", debugger_register_text( exp->types.reg ) );
-    return 0;
-
   case DEBUGGER_EXPRESSION_TYPE_UNARYOP:
     return deparse_unaryop( buffer, length, &( exp->types.unaryop ) );
 
   case DEBUGGER_EXPRESSION_TYPE_BINARYOP:
     return deparse_binaryop( buffer, length, &( exp->types.binaryop ) );
 
+  case DEBUGGER_EXPRESSION_TYPE_SYSVAR:
+    debugger_system_variable_text( buffer, length, exp->types.system_variable );
+    return 0;
+
+  case DEBUGGER_EXPRESSION_TYPE_VARIABLE:
+    snprintf( buffer, length, "$%s", exp->types.variable );
+    return 0;
+
   }
 
   ui_error( UI_ERROR_ERROR, "unknown expression type %d", exp->type );
@@ -379,23 +476,26 @@ deparse_unaryop( char *buffer, size_t length,
 		 const struct unaryop_type *unaryop )
 {
   char *operand_buffer; const char *operation_string = NULL;
-  int brackets;
+  const char *operation_suffix = "";
+  int brackets_possible = 1;
+  int brackets = 0;
 
   int error;
 
-  operand_buffer = malloc( length );
-  if( !operand_buffer ) {
-    ui_error( UI_ERROR_ERROR, "out of memory at %s:%d", __FILE__, __LINE__ );
-    return 1;
-  }
+  operand_buffer = libspectrum_new( char, length );
 
   error = debugger_expression_deparse( operand_buffer, length, unaryop->op );
-  if( error ) { free( operand_buffer ); return error; }
+  if( error ) { libspectrum_free( operand_buffer ); return error; }
 
   switch( unaryop->operation ) {
   case '!': operation_string = "!"; break;
   case '~': operation_string = "~"; break;
   case '-': operation_string = "-"; break;
+  case DEBUGGER_TOKEN_DEREFERENCE:
+    operation_string = "[";
+    operation_suffix = "]";
+    brackets_possible = 0;
+    break;
 
   default:
     ui_error( UI_ERROR_ERROR, "unknown unary operation %d",
@@ -403,14 +503,15 @@ deparse_unaryop( char *buffer, size_t length,
     fuse_abort();
   }
 
-  brackets = ( unaryop->op->precedence                  < 
-	       unaryop_precedence( unaryop->operation )   );
+  if( brackets_possible )
+    brackets = ( unaryop->op->precedence                  < 
+                 unaryop_precedence( unaryop->operation )   );
     
-  snprintf( buffer, length, "%s%s%s%s", operation_string,
+  snprintf( buffer, length, "%s%s%s%s%s", operation_string,
 	    brackets ? "( " : "", operand_buffer,
-	    brackets ? " )" : "" );
+	    brackets ? " )" : "", operation_suffix );
 
-  free( operand_buffer );
+  libspectrum_free( operand_buffer );
 
   return 0;
 }
@@ -424,20 +525,16 @@ deparse_binaryop( char *buffer, size_t length,
 
   int error;
 
-  operand1_buffer = malloc( 2 * length );
-  if( !operand1_buffer ) {
-    ui_error( UI_ERROR_ERROR, "out of memory at %s:%d", __FILE__, __LINE__ );
-    return 1;
-  }
+  operand1_buffer = libspectrum_new( char, 2 * length );
   operand2_buffer = &operand1_buffer[ length ];
 
   error = debugger_expression_deparse( operand1_buffer, length,
 				       binaryop->op1 );
-  if( error ) { free( operand1_buffer ); return error; }
+  if( error ) { libspectrum_free( operand1_buffer ); return error; }
 
   error = debugger_expression_deparse( operand2_buffer, length,
 				       binaryop->op2 );
-  if( error ) { free( operand1_buffer ); return error; }
+  if( error ) { libspectrum_free( operand1_buffer ); return error; }
 
   switch( binaryop->operation ) {
   case    '+': operation_string = "+";  break;
@@ -474,7 +571,7 @@ deparse_binaryop( char *buffer, size_t length,
 	    brackets_necessary2 ? "( " : "", operand2_buffer,
 	    brackets_necessary2 ? " )" : "" );
 
-  free( operand1_buffer );
+  libspectrum_free( operand1_buffer );
 
   return 0;
 }
@@ -535,7 +632,7 @@ is_non_associative( int operation )
   case '+': case '*': return 0;
   case '-': case '/': return 1;
 
-  /* None of the comparision operators are associative due to them
+  /* None of the comparison operators are associative due to them
      returning truth values */
   case DEBUGGER_TOKEN_EQUAL_TO:
   case DEBUGGER_TOKEN_NOT_EQUAL_TO:
