@@ -1,7 +1,5 @@
 /* fuse.c: The Free Unix Spectrum Emulator
-   Copyright (c) 1999-2010 Philip Kendall and others
-
-   $Id: fuse.c 4165 2010-09-30 21:55:05Z pak21 $
+   Copyright (c) 1999-2016 Philip Kendall and others
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -39,41 +37,59 @@
 
 #include <unistd.h>
 
-#ifdef UI_SDL
+/* We need to include SDL.h on Mac O X and Windows to do some magic
+   bootstrapping by redefining main. As we now allow SDL joystick code to be
+   used in the GTK+ and Xlib UIs we need to also do the magic when that code is
+   in use, feel free to look away for the next line */
+#if defined UI_SDL || (defined USE_JOYSTICK && !defined HAVE_JSW_H && (defined UI_X || defined UI_GTK) )
 #include <SDL.h>		/* Needed on MacOS X and Windows */
-#endif				/* #ifdef UI_SDL */
+#endif /* #if defined UI_SDL || (defined USE_JOYSTICK && !defined HAVE_JSW_H && (defined UI_X || defined UI_GTK) ) */
 
 #ifdef GEKKO
 #include <fat.h>
 #endif				/* #ifdef GEKKO */
 
-#include "ay.h"
-#include "dck.h"
+#ifdef HAVE_LIB_XML2
+#include <libxml/encoding.h>
+#endif
+
 #include "debugger/debugger.h"
-#include "disk/beta.h"
-#include "disk/fdd.h"
 #include "display.h"
 #include "event.h"
-#include "fuller.h"
 #include "fuse.h"
-#include "ide/divide.h"
-#include "ide/simpleide.h"
-#include "ide/zxatasp.h"
-#include "ide/zxcf.h"
-#include "if1.h"
-#include "if2.h"
-#include "joystick.h"
-#include "kempmouse.h"
+#include "infrastructure/startup_manager.h"
 #include "keyboard.h"
 #include "machine.h"
-#include "melodik.h"
+#include "machines/machines_periph.h"
 #include "memory.h"
-#include "pokefinder/pokefinder.h"
-#include "printer.h"
+#include "module.h"
+#include "movie.h"
+#include "mempool.h"
+#include "peripherals/ay.h"
+#include "peripherals/dck.h"
+#include "peripherals/disk/beta.h"
+#include "peripherals/disk/didaktik.h"
+#include "peripherals/disk/fdd.h"
+#include "peripherals/fuller.h"
+#include "peripherals/ide/divide.h"
+#include "peripherals/ide/simpleide.h"
+#include "peripherals/ide/zxatasp.h"
+#include "peripherals/ide/zxcf.h"
+#include "peripherals/joystick.h"
+#include "peripherals/if1.h"
+#include "peripherals/if2.h"
+#include "peripherals/kempmouse.h"
+#include "peripherals/melodik.h"
+#include "peripherals/printer.h"
+#include "peripherals/scld.h"
+#include "peripherals/speccyboot.h"
+#include "peripherals/spectranet.h"
+#include "peripherals/ula.h"
+#include "peripherals/usource.h"
+#include "pokefinder/pokemem.h"
 #include "profile.h"
 #include "psg.h"
 #include "rzx.h"
-#include "scld.h"
 #include "settings.h"
 #include "slt.h"
 #include "snapshot.h"
@@ -83,17 +99,14 @@
 #include "timer/timer.h"
 #include "ui/scaler/scaler.h"
 #include "ui/ui.h"
-#include "ula.h"
+#include "ui/uimedia.h"
 #include "unittests/unittests.h"
 #include "utils.h"
 
 #include "z80/z80.h"
 
 /* What name were we called under? */
-char *fuse_progname;
-
-/* Which directory were we started in? */
-char fuse_directory[ PATH_MAX ];
+const char *fuse_progname;
 
 /* A flag to say when we want to exit the emulator */
 int fuse_exiting;
@@ -105,7 +118,7 @@ int fuse_emulation_paused;
 libspectrum_creator *fuse_creator;
 
 /* The earliest version of libspectrum we need */
-static const char *LIBSPECTRUM_MIN_VERSION = "0.5.0";
+static const char * const LIBSPECTRUM_MIN_VERSION = "0.5.0";
 
 /* The various types of file we may want to run on startup */
 typedef struct start_files_t {
@@ -114,6 +127,8 @@ typedef struct start_files_t {
   const char *disk_opus;
   const char *disk_plusd;
   const char *disk_beta;
+  const char *disk_didaktik80;
+  const char *disk_disciple;
   const char *dock;
   const char *if2;
   const char *playback;
@@ -129,9 +144,13 @@ typedef struct start_files_t {
 
 } start_files_t;
 
+/* Context for the display startup routine */
+static display_startup_context display_context;
+
 static int fuse_init(int argc, char **argv);
 
-static int creator_init( void );
+static void creator_register_startup( void );
+
 static void fuse_show_copyright(void);
 static void fuse_show_version( void );
 static void fuse_show_help( void );
@@ -183,7 +202,138 @@ int main(int argc, char **argv)
 
 }
 
-#include "mempool.h"
+static int
+fuse_libspectrum_init( void *context )
+{
+  if( libspectrum_check_version( LIBSPECTRUM_MIN_VERSION ) ) {
+    if( libspectrum_init() ) return 1;
+  } else {
+    ui_error( UI_ERROR_ERROR,
+              "libspectrum version %s found, but %s required",
+	      libspectrum_version(), LIBSPECTRUM_MIN_VERSION );
+    return 1;
+  }
+
+  return 0;
+}
+
+static void
+libspectrum_register_startup( void )
+{
+  startup_manager_module dependencies[] = {
+    STARTUP_MANAGER_MODULE_DISPLAY
+  };
+  startup_manager_register( STARTUP_MANAGER_MODULE_LIBSPECTRUM, dependencies,
+                            ARRAY_SIZE( dependencies ), fuse_libspectrum_init,
+                            NULL, NULL );
+}
+
+static int
+libxml2_init( void *context )
+{
+#ifdef HAVE_LIB_XML2
+  LIBXML_TEST_VERSION
+#endif
+
+  return 0;
+}
+
+static void
+libxml2_register_startup( void )
+{
+  startup_manager_module dependencies[] = { STARTUP_MANAGER_MODULE_SETUID };
+  startup_manager_register( STARTUP_MANAGER_MODULE_LIBXML2, dependencies,
+                            ARRAY_SIZE( dependencies ), libxml2_init, NULL,
+                            NULL );
+}
+
+static int
+setuid_init( void *context )
+{
+#ifdef HAVE_GETEUID
+  int error;
+
+  /* Drop root privs if we have them */
+  if( !geteuid() ) {
+    error = setuid( getuid() );
+    if( error ) {
+      ui_error( UI_ERROR_ERROR, "Could not drop root privileges" );
+      return 1;
+    }
+  }
+#endif				/* #ifdef HAVE_GETEUID */
+
+  return 0;
+}
+
+static void
+setuid_register_startup()
+{
+  startup_manager_module dependencies[] = {
+    STARTUP_MANAGER_MODULE_DISPLAY,
+    STARTUP_MANAGER_MODULE_LIBSPECTRUM,
+  };
+  startup_manager_register( STARTUP_MANAGER_MODULE_SETUID, dependencies,
+                            ARRAY_SIZE( dependencies ), setuid_init, NULL,
+                            NULL );
+}
+
+static int
+run_startup_manager( int *argc, char ***argv )
+{
+  startup_manager_init();
+
+  display_context.argc = argc;
+  display_context.argv = argv;
+
+  /* Get every module to register its init function */
+  ay_register_startup();
+  beta_register_startup();
+  creator_register_startup();
+  debugger_register_startup();
+  didaktik80_register_startup();
+  disciple_register_startup();
+  display_register_startup( &display_context );
+  divide_register_startup();
+  event_register_startup();
+  fdd_register_startup();
+  fuller_register_startup();
+  if1_register_startup();
+  if2_register_startup();
+  kempmouse_register_startup();
+  libspectrum_register_startup();
+  libxml2_register_startup();
+  machine_register_startup();
+  machines_periph_register_startup();
+  melodik_register_startup();
+  memory_register_startup();
+  mempool_register_startup();
+  opus_register_startup();
+  plusd_register_startup();
+  printer_register_startup();
+  profile_register_startup();
+  psg_register_startup();
+  rzx_register_startup();
+  scld_register_startup();
+  settings_register_startup();
+  setuid_register_startup();
+  simpleide_register_startup();
+  slt_register_startup();
+  sound_register_startup();
+  speccyboot_register_startup();
+  specdrum_register_startup();
+  spectranet_register_startup();
+  spectrum_register_startup();
+  tape_register_startup();
+  timer_register_startup();
+  ula_register_startup();
+  usource_register_startup();
+  z80_register_startup();
+  zxatasp_register_startup();
+  zxcf_register_startup();
+
+  return startup_manager_run();
+}
 
 static int fuse_init(int argc, char **argv)
 {
@@ -209,13 +359,6 @@ static int fuse_init(int argc, char **argv)
   if( display_init(&argc,&argv) ) return 1;
 #endif
 
-  if( !getcwd( fuse_directory, PATH_MAX - 1 ) ) {
-    ui_error( UI_ERROR_ERROR, "error getting current working directory: %s",
-	      strerror( errno ) );
-    return 1;
-  }
-  strcat( fuse_directory, FUSE_DIR_SEP_STR );
-
   if( settings_init( &first_arg, argc, argv ) ) return 1;
 
   if( settings_current.show_version ) {
@@ -226,94 +369,19 @@ static int fuse_init(int argc, char **argv)
     return 0;
   }
 
-  start_scaler = strdup( settings_current.start_scaler_mode );
-  if( !start_scaler ) {
-    ui_error( UI_ERROR_ERROR, "Out of memory at %s:%d", __FILE__, __LINE__ );
-    return 1;
-  }
+  start_scaler = utils_safe_strdup( settings_current.start_scaler_mode );
 
-  /* Windows will create a console for our output if there isn't one already,
-   * so we don't display the copyright message on Win32. */
-#ifndef WIN32
   fuse_show_copyright();
-#endif
 
-  /* FIXME: order of these initialisation calls. Work out what depends on
-     what */
-  /* FIXME FIXME 20030407: really do this soon. This is getting *far* too
-     hairy */
-  fuse_joystick_init ();
+  fuse_joystick_init();
   fuse_keyboard_init();
 
-  if( event_init() ) return 1;
-  
-#ifndef GEKKO
-  if( display_init(&argc,&argv) ) return 1;
-#endif
-
-  if( libspectrum_check_version( LIBSPECTRUM_MIN_VERSION ) ) {
-    if( libspectrum_init() ) return 1;
-  } else {
-    ui_error( UI_ERROR_ERROR,
-              "libspectrum version %s found, but %s required",
-	      libspectrum_version(), LIBSPECTRUM_MIN_VERSION );
-    return 1;
-  }
-
-  /* Must be called after libspectrum_init() so we can get the gcrypt
-     version */
-  if( creator_init() ) return 1;
-
-#ifdef HAVE_GETEUID
-  /* Drop root privs if we have them */
-  if( !geteuid() ) { setuid( getuid() ); }
-#endif				/* #ifdef HAVE_GETEUID */
-
-  if( mempool_init() ) return 1;
-  if( memory_init() ) return 1;
-
-  if( debugger_init() ) return 1;
-
-  if( spectrum_init() ) return 1;
-  if( printer_init() ) return 1;
-  if( rzx_init() ) return 1;
-  if( psg_init() ) return 1;
-  if( beta_init() ) return 1;
-  if( opus_init() ) return 1;
-  if( plusd_init() ) return 1;
-  if( fdd_init_events() ) return 1;
-  if( simpleide_init() ) return 1;
-  if( zxatasp_init() ) return 1;
-  if( zxcf_init() ) return 1;
-  if( if1_init() ) return 1;
-  if( if2_init() ) return 1;
-  if( divide_init() ) return 1;
-  if( scld_init() ) return 1;
-  if( ula_init() ) return 1;
-  if( ay_init() ) return 1;
-  if( slt_init() ) return 1;
-  if( profile_init() ) return 1;
-  if( kempmouse_init() ) return 1;
-  if( fuller_init() ) return 1;
-  if( melodik_init() ) return 1;
-
-  error = pokefinder_clear(); if( error ) return error;
-
-  if( z80_init() ) return 1;
-
-  if( timer_init() ) return 1;
-
-  error = timer_estimate_reset(); if( error ) return error;
-
-  error = machine_init_machines();
-  if( error ) return error;
+  if( run_startup_manager( &argc, &argv ) ) return 1;
 
   error = machine_select_id( settings_current.start_machine );
   if( error ) return error;
 
-  error = tape_init(); if( error ) return error;
-
-  error = scaler_select_id( start_scaler ); free( start_scaler );
+  error = scaler_select_id( start_scaler ); libspectrum_free( start_scaler );
   if( error ) return error;
 
   if( setup_start_files( &start_files ) ) return 1;
@@ -326,12 +394,13 @@ static int fuse_init(int argc, char **argv)
   if( ui_mouse_present ) ui_mouse_grabbed = ui_mouse_grab( 1 );
 
   fuse_emulation_paused = 0;
+  movie_init();
 
   return 0;
 }
 
-static
-int creator_init( void )
+static int
+creator_init( void *context )
 {
   size_t i;
   unsigned int version[4] = { 0, 0, 0, 0 };
@@ -363,12 +432,7 @@ int creator_init( void )
 					 version[2] * 0x100 + version[3] );
   if( error ) { libspectrum_creator_free( fuse_creator ); return error; }
 
-  custom = malloc( CUSTOM_SIZE );
-  if( !custom ) {
-    ui_error( UI_ERROR_ERROR, "out of memory at %s:%d", __FILE__, __LINE__ );
-    libspectrum_creator_free( fuse_creator );
-    return 1;
-  }
+  custom = libspectrum_new( char, CUSTOM_SIZE );
 
   gcrypt_version = libspectrum_gcrypt_version();
   if( !gcrypt_version ) gcrypt_version = "not available";
@@ -381,23 +445,38 @@ int creator_init( void )
     fuse_creator, (libspectrum_byte*)custom, strlen( custom )
   );
   if( error ) {
-    free( custom ); libspectrum_creator_free( fuse_creator );
+    libspectrum_free( custom ); libspectrum_creator_free( fuse_creator );
     return error;
   }
 
   return 0;
 }
 
+static void
+creator_end( void )
+{
+  libspectrum_creator_free( fuse_creator );
+}
+
+static void
+creator_register_startup( void )
+{
+  startup_manager_module dependencies[] = { STARTUP_MANAGER_MODULE_SETUID };
+  startup_manager_register( STARTUP_MANAGER_MODULE_CREATOR, dependencies,
+                            ARRAY_SIZE( dependencies ), creator_init, NULL,
+                            creator_end );
+}
+
 static void fuse_show_copyright(void)
 {
   printf( "\n" );
   fuse_show_version();
   printf(
-   "Copyright (c) 1999-2010 Philip Kendall and others; see the file\n"
+   FUSE_COPYRIGHT "; see the file\n"
    "'AUTHORS' for more details.\n"
    "\n"
    "For help, please mail <fuse-emulator-devel@lists.sf.net> or use\n"
-   "the forums at <http://sourceforge.net/forum/?group_id=91293>.\n"
+   "the forums at <http://sourceforge.net/p/fuse-emulator/discussion/>.\n"
    "\n"
    "This program is distributed in the hope that it will be useful,\n"
    "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
@@ -418,13 +497,10 @@ static void fuse_show_help( void )
    "\nAvailable command-line options:\n\n"
    "Boolean options (use `--no-<option>' to turn off):\n\n"
    "--auto-load            Automatically load tape files when opened.\n"
-   "--beeper-stereo        Add fake stereo to beeper emulation.\n"
    "--compress-rzx         Write RZX files out compressed.\n"
-   "--double-screen        Write screenshots out as double size.\n"
    "--issue2               Emulate an Issue 2 Spectrum.\n"
    "--kempston             Emulate the Kempston joystick on QAOP<space>.\n"
    "--loading-sound        Emulate the sound of tapes loading.\n"
-   "--separation           Use ACB stereo for the AY-3-8912 sound chip.\n"
    "--sound                Produce sound.\n"
    "--sound-force-8bit     Generate 8-bit sound even if 16-bit is available.\n"
    "--slt                  Turn SLT traps on.\n"
@@ -434,11 +510,16 @@ static void fuse_show_help( void )
    "--machine <type>       Which machine should be emulated?\n"
    "--playback <filename>  Play back RZX file <filename>.\n"
    "--record <filename>    Record to RZX file <filename>.\n"
+   "--separation <type>    Use ACB/ABC stereo for the AY-3-8912 sound chip.\n"
    "--snapshot <filename>  Load snapshot <filename>.\n"
    "--speed <percentage>   How fast should emulation run?\n"
    "--fb-mode <mode>       Which mode should be used for FB?\n"
    "--tape <filename>      Open tape file <filename>.\n"
-   "--version              Print version number and exit.\n\n" );
+   "--version              Print version number and exit.\n"
+   "\n"
+   "For help, please mail <fuse-emulator-devel@lists.sf.net> or use\n"
+   "the forums at <http://sourceforge.net/p/fuse-emulator/discussion/>.\n"
+   "For complete documentation, see the manual page of Fuse.\n\n" );
 }
 
 /* Stop all activities associated with actual Spectrum emulation */
@@ -487,6 +568,8 @@ setup_start_files( start_files_t *start_files )
   start_files->disk_plus3 = settings_current.plus3disk_file;
   start_files->disk_opus = settings_current.opusdisk_file;
   start_files->disk_plusd = settings_current.plusddisk_file;
+  start_files->disk_didaktik80 = settings_current.didaktik80disk_file;
+  start_files->disk_disciple = settings_current.discipledisk_file;
   start_files->disk_beta = settings_current.betadisk_file;
   start_files->dock = settings_current.dck_file;
   start_files->if2 = settings_current.if2_file;
@@ -495,16 +578,22 @@ setup_start_files( start_files_t *start_files )
   start_files->snapshot = settings_current.snapshot;
   start_files->tape = settings_current.tape_file;
 
-  start_files->simpleide_master = settings_current.simpleide_master_file;
-  start_files->simpleide_slave  = settings_current.simpleide_slave_file;
+  start_files->simpleide_master =
+    utils_safe_strdup( settings_current.simpleide_master_file );
+  start_files->simpleide_slave = 
+    utils_safe_strdup( settings_current.simpleide_slave_file );
 
-  start_files->zxatasp_master = settings_current.zxatasp_master_file;
-  start_files->zxatasp_slave  = settings_current.zxatasp_slave_file;
+  start_files->zxatasp_master =
+    utils_safe_strdup( settings_current.zxatasp_master_file );
+  start_files->zxatasp_slave =
+    utils_safe_strdup( settings_current.zxatasp_slave_file );
 
-  start_files->zxcf = settings_current.zxcf_pri_file;
+  start_files->zxcf = utils_safe_strdup( settings_current.zxcf_pri_file );
 
-  start_files->divide_master = settings_current.divide_master_file;
-  start_files->divide_slave  = settings_current.divide_slave_file;
+  start_files->divide_master =
+    utils_safe_strdup( settings_current.divide_master_file );
+  start_files->divide_slave =
+    utils_safe_strdup( settings_current.divide_slave_file );
 
   start_files->mdr[0] = settings_current.mdr_file;
   start_files->mdr[1] = settings_current.mdr_file2;
@@ -544,7 +633,10 @@ parse_nonoption_args( int argc, char **argv, int first_arg,
 
     error = libspectrum_identify_file_with_class( &type, &class, filename,
 						  file.buffer, file.length );
-    if( error ) return error;
+    if( error ) {
+      utils_close_file( &file );
+      return error;
+    }
 
     switch( class ) {
 
@@ -576,27 +668,36 @@ parse_nonoption_args( int argc, char **argv, int first_arg,
     case LIBSPECTRUM_CLASS_DISK_OPUS:
       start_files->disk_opus = filename; break;
 
+    case LIBSPECTRUM_CLASS_DISK_DIDAKTIK:
+      start_files->disk_didaktik80 = filename; break;
+
     case LIBSPECTRUM_CLASS_DISK_PLUSD:
-      start_files->disk_plusd = filename; break;
+      if( periph_is_active( PERIPH_TYPE_DISCIPLE ) )
+        start_files->disk_disciple = filename;
+      else
+        start_files->disk_plusd = filename;
+      break;
 
     case LIBSPECTRUM_CLASS_DISK_TRDOS:
       start_files->disk_beta = filename; break;
 
     case LIBSPECTRUM_CLASS_DISK_GENERIC:
-      if( machine_current->machine == LIBSPECTRUM_MACHINE_PLUS3 ||
-          machine_current->machine == LIBSPECTRUM_MACHINE_PLUS2A )
+      if( machine_current->capabilities &
+                 LIBSPECTRUM_MACHINE_CAPABILITY_PLUS3_DISK )
         start_files->disk_plus3 = filename;
-      else if( machine_current->machine == LIBSPECTRUM_MACHINE_PENT ||
-          machine_current->machine == LIBSPECTRUM_MACHINE_PENT512 ||
-          machine_current->machine == LIBSPECTRUM_MACHINE_PENT1024 ||
-          machine_current->machine == LIBSPECTRUM_MACHINE_SCORP )
+      else if( machine_current->capabilities &
+                 LIBSPECTRUM_MACHINE_CAPABILITY_TRDOS_DISK )
         start_files->disk_beta = filename; 
       else {
-        if( periph_beta128_active )
+        if( periph_is_active( PERIPH_TYPE_BETA128 ) )
           start_files->disk_beta = filename; 
-        else if( periph_plusd_active )
+        else if( periph_is_active( PERIPH_TYPE_PLUSD ) )
           start_files->disk_plusd = filename;
-        else if( periph_opus_active )
+        else if( periph_is_active( PERIPH_TYPE_DIDAKTIK80 ) )
+          start_files->disk_didaktik80 = filename;
+        else if( periph_is_active( PERIPH_TYPE_DISCIPLE ) )
+          start_files->disk_disciple = filename;
+        else if( periph_is_active( PERIPH_TYPE_OPUS ) )
           start_files->disk_opus = filename;
       }
       break;
@@ -619,6 +720,12 @@ parse_nonoption_args( int argc, char **argv, int first_arg,
     case LIBSPECTRUM_CLASS_TAPE:
       start_files->tape = filename; break;
 
+    case LIBSPECTRUM_CLASS_AUXILIARY:
+      if( type == LIBSPECTRUM_ID_AUX_POK ) {
+        pokemem_set_pokfile( filename );
+      }
+      break;
+
     case LIBSPECTRUM_CLASS_UNKNOWN:
       ui_error( UI_ERROR_WARNING, "couldn't identify '%s'; ignoring it",
 		filename );
@@ -630,6 +737,8 @@ parse_nonoption_args( int argc, char **argv, int first_arg,
       break;
 
     }
+
+    utils_close_file( &file );
   }
 
   return 0;
@@ -668,12 +777,12 @@ do_start_files( start_files_t *start_files )
     start_files->dock = NULL;
   }
 
-  /* Can't use disks and the Interface II simultaneously */
+  /* Can't use disks and the Interface 2 simultaneously */
   if( ( start_files->disk_plus3 || start_files->disk_beta ) &&
       start_files->if2                                          ) {
     ui_error(
       UI_ERROR_WARNING,
-      "can't use disks and the Interface II simultaneously; cartridge ignored"
+      "can't use disks and the Interface 2 simultaneously; cartridge ignored"
     );
     start_files->if2 = NULL;
   }
@@ -697,6 +806,16 @@ do_start_files( start_files_t *start_files )
     if( error ) return error;
   }
 
+  if( start_files->disk_didaktik80 ) {
+    error = utils_open_file( start_files->disk_didaktik80, autoload, NULL );
+    if( error ) return error;
+  }
+
+  if( start_files->disk_disciple ) {
+    error = utils_open_file( start_files->disk_disciple, autoload, NULL );
+    if( error ) return error;
+  }
+
   if( start_files->disk_opus ) {
     error = utils_open_file( start_files->disk_opus, autoload, NULL );
     if( error ) return error;
@@ -800,35 +919,21 @@ do_start_files( start_files_t *start_files )
 /* Tidy-up function called at end of emulation */
 static int fuse_end(void)
 {
-  /* Must happen before memory is deallocated as we read the character
-     set from memory for the text output */
-  printer_end();
+  movie_stop();		/* stop movie recording */
 
-  /* also required before memory is deallocated on Fuse for OS X where
-     settings need to look up machine names etc. */
-  settings_end();
+  startup_manager_run_end();
 
-  psg_end();
-  rzx_end();
-  debugger_end();
-  simpleide_end();
-  zxatasp_end();
-  zxcf_end();
-  if1_end();
-  divide_end();
-  opus_end();
-  plusd_end();
-
-  machine_end();
-
-  timer_end();
-
-  sound_end();
-  event_end();
+  periph_end();
+  fuse_keyboard_end();
   fuse_joystick_end();
   ui_end();
+  ui_media_drive_end();
+  module_end();
+  pokemem_end();
 
-  libspectrum_creator_free( fuse_creator );
+  svg_capture_end();
+
+  libspectrum_end();
 
   return 0;
 }
