@@ -1,7 +1,5 @@
 /* keyboard.c: Routines for dealing with the Spectrum's keyboard
-   Copyright (c) 1999-2000 Philip Kendall
-
-   $Id: keyboard.c 3265 2007-10-28 12:35:45Z pak21 $
+   Copyright (c) 1999-2013 Philip Kendall
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -33,7 +31,6 @@
 
 #include "ui/ui.h"
 #include "keyboard.h"
-#include "scld.h"
 
 /* Bit masks for each of the eight keyboard half-rows; `AND' the selected
    ones of these to get the value to return
@@ -107,6 +104,14 @@ static struct spectrum_keys_wrapper spectrum_keys_table[] = {
   { INPUT_KEY_slash,       { KEYBOARD_v,     KEYBOARD_Symbol } },
   { INPUT_KEY_Shift_R,     { KEYBOARD_NONE,  KEYBOARD_Caps   } },
 
+  { INPUT_KEY_asterisk,    { KEYBOARD_b,     KEYBOARD_Symbol } },
+  { INPUT_KEY_dollar,      { KEYBOARD_4,     KEYBOARD_Symbol } },
+  { INPUT_KEY_exclam,      { KEYBOARD_1,     KEYBOARD_Symbol } },
+  { INPUT_KEY_less,        { KEYBOARD_r,     KEYBOARD_Symbol } },
+  { INPUT_KEY_parenright,  { KEYBOARD_9,     KEYBOARD_Symbol } },
+  { INPUT_KEY_colon,       { KEYBOARD_z,     KEYBOARD_Symbol } },
+  { INPUT_KEY_plus,        { KEYBOARD_k,     KEYBOARD_Symbol } },
+
   { INPUT_KEY_Control_L,   { KEYBOARD_NONE,  KEYBOARD_Symbol } },
   { INPUT_KEY_Alt_L,       { KEYBOARD_NONE,  KEYBOARD_Symbol } },
   { INPUT_KEY_Meta_L,      { KEYBOARD_NONE,  KEYBOARD_Symbol } },
@@ -125,6 +130,8 @@ static struct spectrum_keys_wrapper spectrum_keys_table[] = {
   { INPUT_KEY_Up,          { KEYBOARD_7,     KEYBOARD_Caps   } },
   { INPUT_KEY_Right,       { KEYBOARD_8,     KEYBOARD_Caps   } },
 
+  { INPUT_KEY_KP_Enter,    { KEYBOARD_Enter, KEYBOARD_NONE   } },
+
   { INPUT_KEY_NONE, { KEYBOARD_NONE, KEYBOARD_NONE } } /* End marker */
 
 };
@@ -253,7 +260,7 @@ struct key_text_t key_text_table[] = {
 
   { KEYBOARD_JOYSTICK_FIRE, "Joystick Fire" },
 
-  { -1, NULL },		/* End marker */
+  { KEYBOARD_NONE, NULL },		/* End marker */
 
 };
 
@@ -288,11 +295,19 @@ void fuse_keyboard_init(void)
 
   key_text = g_hash_table_new( g_int_hash, g_int_equal );
 
-  for( ptr4 = key_text_table; ptr4->key != -1; ptr4++ )
+  for( ptr4 = key_text_table; ptr4->text != NULL; ptr4++ )
     g_hash_table_insert( key_text, &( ptr4->key ), &( ptr4->text ) );
 
 }
 
+void fuse_keyboard_end(void)
+{
+  g_hash_table_destroy( keyboard_data );
+  g_hash_table_destroy( spectrum_keys );
+  g_hash_table_destroy( keysyms_hash );
+  g_hash_table_destroy( key_text );
+}
+
 libspectrum_byte
 keyboard_read( libspectrum_byte porth )
 {
