@@ -1,7 +1,5 @@
 /* scorpion.c: Scorpion 256K specific routines
-   Copyright (c) 1999-2007 Philip Kendall, Fredrick Meunier and Stuart Brady
-
-   $Id: scorpion.c 4099 2009-10-22 10:59:02Z fredm $
+   Copyright (c) 1999-2011 Philip Kendall, Fredrick Meunier and Stuart Brady
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,47 +28,26 @@
 
 #include <libspectrum.h>
 
-#include "ay.h"
 #include "compat.h"
-#include "disk/beta.h"
-#include "joystick.h"
 #include "machine.h"
 #include "machines.h"
+#include "machines_periph.h"
 #include "memory.h"
 #include "pentagon.h"
-#include "printer.h"
+#include "peripherals/disk/beta.h"
 #include "settings.h"
 #include "scorpion.h"
 #include "spec128.h"
 #include "spec48.h"
 #include "specplus3.h"
 #include "spectrum.h"
-#include "ula.h"
 
 static int scorpion_reset( void );
 static int scorpion_memory_map( void );
 
-static const periph_t peripherals[] = {
-  { 0x00ff, 0x001f, pentagon_select_1f_read, beta_cr_write },
-  { 0x00ff, 0x003f, beta_tr_read, beta_tr_write },
-  { 0x00ff, 0x005f, beta_sec_read, beta_sec_write },
-  { 0x00ff, 0x007f, beta_dr_read, beta_dr_write },
-  { 0x00ff, 0x00fe, ula_read, ula_write },
-  { 0x00ff, 0x00ff, beta_sp_read, beta_sp_write },
-  { 0xc002, 0xc000, ay_registerport_read, ay_registerport_write },
-  { 0xc002, 0x8000, NULL, ay_dataport_write },
-  { 0xc002, 0x4000, NULL, spec128_memoryport_write },
-  { 0xf002, 0x1000, NULL, specplus3_memoryport2_write },
-};
-
-static const size_t peripherals_count =
-  sizeof( peripherals ) / sizeof( periph_t );
-
 int
 scorpion_init( fuse_machine_info *machine )
 {
-  int i;
-
   machine->machine = LIBSPECTRUM_MACHINE_SCORP;
   machine->id = "scorpion";
 
@@ -80,13 +57,13 @@ scorpion_init( fuse_machine_info *machine )
   machine->ram.port_from_ula  = pentagon_port_from_ula;
   machine->ram.contend_delay  = spectrum_contend_delay_none;
   machine->ram.contend_delay_no_mreq = spectrum_contend_delay_none;
+  machine->ram.valid_pages    = 16;
 
   machine->unattached_port = spectrum_unattached_port_none;
 
   machine->shutdown = NULL;
 
   machine->memory_map = scorpion_memory_map;
-  for( i = 0; i < 2; i++ ) beta_memory_map_romcs[i].bank = MEMORY_BANK_ROMCS;
 
   return 0;
 }
@@ -94,18 +71,18 @@ scorpion_init( fuse_machine_info *machine )
 int
 scorpion_reset(void)
 {
-  int i, error;
+  int error;
 
-  error = machine_load_rom( 0, 0, settings_current.rom_scorpion_0,
+  error = machine_load_rom( 0, settings_current.rom_scorpion_0,
                             settings_default.rom_scorpion_0, 0x4000 );
   if( error ) return error;
-  error = machine_load_rom( 2, 1, settings_current.rom_scorpion_1,
+  error = machine_load_rom( 1, settings_current.rom_scorpion_1,
                             settings_default.rom_scorpion_1, 0x4000 );
   if( error ) return error;
-  error = machine_load_rom( 4, 2, settings_current.rom_scorpion_2,
+  error = machine_load_rom( 2, settings_current.rom_scorpion_2,
                             settings_default.rom_scorpion_2, 0x4000 );
   if( error ) return error;
-  error = machine_load_rom_bank( beta_memory_map_romcs, 0, 0,
+  error = machine_load_rom_bank( beta_memory_map_romcs, 0,
                                  settings_current.rom_scorpion_3,
                                  settings_default.rom_scorpion_3, 0x4000 );
   if( error ) return error;
@@ -116,14 +93,16 @@ scorpion_reset(void)
   machine_current->ram.last_byte2 = 0;
   machine_current->ram.special = 0;
 
-  /* Mark the second 128K as present/writeable */
-  for( i = 16; i < 32; i++ )
-    memory_map_ram[i].writable = 1;
+  periph_clear();
+  machines_periph_pentagon();
+
+  /* +3-style memory paging */
+  periph_set_present( PERIPH_TYPE_128_MEMORY, PERIPH_PRESENT_NEVER );
+  periph_set_present( PERIPH_TYPE_PLUS3_MEMORY, PERIPH_PRESENT_ALWAYS );
+
+  /* Later style Betadisk 128 interface */
+  periph_set_present( PERIPH_TYPE_BETA128_PENTAGON_LATE, PERIPH_PRESENT_ALWAYS );
 
-  error = periph_setup( peripherals, peripherals_count );
-  if( error ) return error;
-  periph_setup_kempston( PERIPH_PRESENT_OPTIONAL );
-  periph_setup_beta128( PERIPH_PRESENT_ALWAYS );
   periph_update();
 
   beta_builtin = 1;
@@ -138,7 +117,6 @@ static int
 scorpion_memory_map( void )
 {
   int rom, page, screen;
-  size_t i;
 
   screen = ( machine_current->ram.last_byte & 0x08 ) ? 7 : 5;
   if( memory_current_screen != screen ) {
@@ -155,8 +133,7 @@ scorpion_memory_map( void )
   machine_current->ram.current_rom = rom;
 
   if( machine_current->ram.last_byte2 & 0x01 ) {
-    memory_map_home[0] = &memory_map_ram[ 0 ];
-    memory_map_home[1] = &memory_map_ram[ 1 ];
+    memory_map_16k( 0x0000, memory_map_ram, 0 );
     machine_current->ram.special = 1;
   } else {
     spec128_select_rom( rom );
@@ -168,9 +145,6 @@ scorpion_memory_map( void )
   spec128_select_page( page );
   machine_current->ram.current_page = page;
 
-  for( i = 0; i < 8; i++ )
-    memory_map_read[i] = memory_map_write[i] = *memory_map_home[i];
-
   memory_romcs_map();
 
   return 0;
