@@ -1,9 +1,7 @@
 /* sdlsound.c: SDL sound I/O
-   Copyright (c) 2002-2004 Alexander Yurchenko, Russell Marks, Philip Kendall,
+   Copyright (c) 2002-2015 Alexander Yurchenko, Russell Marks, Philip Kendall,
 			   Fredrick Meunier
 
-   $Id: sdlsound.c 3302 2007-11-17 23:38:47Z fredm $
-
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
@@ -60,11 +58,12 @@ sound_lowlevel_init( const char *device, int *freqptr, int *stereoptr )
   /* I'd rather just use setenv, but Windows doesn't have it */
   if( device ) {
     const char *environment = "SDL_AUDIODRIVER=";
-    char *command = malloc( strlen( environment ) + strlen( device ) + 1 );
+    char *command = libspectrum_new( char, strlen( environment ) +
+                                           strlen( device ) + 1 );
     strcpy( command, environment );
     strcat( command, device );
     error = putenv( command );
-    free( command );
+    libspectrum_free( command );
     if( error ) { 
       settings_current.sound = 0;
       ui_error( UI_ERROR_ERROR, "Couldn't set SDL_AUDIODRIVER: %s",
@@ -83,14 +82,18 @@ sound_lowlevel_init( const char *device, int *freqptr, int *stereoptr )
   requested.format = AUDIO_S16SYS;
   requested.callback = sdlwrite;
 
-  hz = (float)machine_current->timings.processor_speed /
+  /* Adjust relative processor speed to deal with adjusting sound generation
+     frequency against emulation speed (more flexible than adjusting generated
+     sample rate) */
+  hz = (float)sound_get_effective_processor_speed() /
               machine_current->timings.tstates_per_frame;
+  /* Amount of audio data we will accumulate before yielding back to the OS.
+     Not much point having more than 100Hz playback, we probably get
+     downgraded by the OS as being a hog too (unlimited Hz limits playback
+     speed to about 2000% on my Mac, 100Hz allows up to 5000% for me) */
+  if( hz > 100.0 ) hz = 100.0;
   sound_framesiz = *freqptr / hz;
-#if !defined AMIGA && !defined __MORPHOS__
-  requested.samples = pow(2.0, floor(log2(sound_framesiz)));
-#else			/* #if !defined AMIGA && !defined __MORPHOS__ */
   requested.samples = sound_framesiz;
-#endif			/* #if !defined AMIGA && !defined __MORPHOS__ */
 
   if ( SDL_OpenAudio( &requested, &received ) < 0 ) {
     settings_current.sound = 0;
@@ -107,15 +110,8 @@ sound_lowlevel_init( const char *device, int *freqptr, int *stereoptr )
     SDL_CloseAudio();
 
     requested.freq = *freqptr;
-
-    hz = (float)machine_current->timings.processor_speed /
-                machine_current->timings.tstates_per_frame;
     sound_framesiz = *freqptr / hz;
-#if !defined AMIGA && !defined __MORPHOS__   
-    requested.samples = pow(2.0, floor(log2(sound_framesiz)));
-#else			/* #if !defined AMIGA && !defined __MORPHOS__ */
     requested.samples = sound_framesiz;
-#endif			/* #if !defined AMIGA && !defined __MORPHOS__ */
 
     if( SDL_OpenAudio( &requested, NULL ) < 0 ) {
       settings_current.sound = 0;
@@ -127,8 +123,6 @@ sound_lowlevel_init( const char *device, int *freqptr, int *stereoptr )
     *stereoptr = received.channels == 1 ? 0 : 1;
   }
 
-  hz = (float)machine_current->timings.processor_speed /
-              machine_current->timings.tstates_per_frame;
   sound_framesiz = *freqptr / hz;
   sound_framesiz <<= 1;
 
@@ -199,7 +193,7 @@ sdlwrite( void *userdata, Uint8 *stream, int len )
 
   /* Try to only read an even number of bytes so as not to fragment a sample */
   len = MIN( len, sfifo_used( &sound_fifo ) );
-  len &= sound_stereo ? 0xfffc : 0xfffe;
+  len &= sound_stereo_ay ? 0xfffc : 0xfffe;
 
   /* Read input_size bytes from fifo into sound stream */
   while( ( f = sfifo_read( &sound_fifo, stream, len ) ) > 0 ) {
