@@ -1,8 +1,7 @@
 /* filesel.c: File selection dialog box
-   Copyright (c) 2001-2005 Matan Ziv-Av, Philip Kendall, Russell Marks,
+   Copyright (c) 2001-2015 Matan Ziv-Av, Philip Kendall, Russell Marks,
 			   Marek Januszewski
-
-   $Id: filesel.c 4246 2011-01-09 20:47:53Z pak21 $
+   Copyright (c) 2015 Sergio Baldoví
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -82,7 +81,7 @@ static int is_drivesel = 0;
 static int is_rootdir;
 #endif				/* #ifdef WIN32 */
 
-#define PAGESIZE (is_saving ? 32 : 36)
+#define ENTRIES_PER_SCREEN (is_saving ? 32 : 36)
 
 /* The number of the filename in the top-left corner of the current
    display, that of the filename which the `cursor' is on, and that
@@ -92,7 +91,8 @@ static size_t top_left_file, current_file, new_current_file;
 static char *widget_get_filename( const char *title, int saving );
 
 static int widget_add_filename( int *allocated, int *number,
-				struct widget_dirent ***namelist, char *name );
+                                struct widget_dirent ***namelist,
+                                const char *name );
 static void widget_scan( char *dir );
 static int widget_select_file( const char *name );
 static int widget_scan_compare( const widget_dirent **a,
@@ -122,7 +122,6 @@ static char *
 widget_get_filename( const char *title, int saving )
 {
   char *filename = NULL;
-  widget_type wtype;
 
   widget_filesel_data data;
 
@@ -130,16 +129,12 @@ widget_get_filename( const char *title, int saving )
   data.title = title;
 
   if( saving ) {
-    wtype = WIDGET_TYPE_FILESELECTOR_SAVE;
+    widget_do_fileselector_save( &data );
   } else {
-    wtype = WIDGET_TYPE_FILESELECTOR;
-  }
-  widget_do( wtype, &data );
-  if( widget_filesel_name ) {
-    filename = strdup( widget_filesel_name );
-    if( !filename )
-      ui_error( UI_ERROR_ERROR, "Out of memory at %s:%d", __FILE__, __LINE__ );
+    widget_do_fileselector( &data );
   }
+  if( widget_filesel_name )
+    filename = utils_safe_strdup( widget_filesel_name );
 
   return filename;
   
@@ -166,8 +161,8 @@ ui_get_save_filename( const char *title )
 }
 
 static int widget_add_filename( int *allocated, int *number,
-				struct widget_dirent ***namelist,
-				char *name ) {
+                                struct widget_dirent ***namelist,
+                                const char *name ) {
   int i; size_t length;
 
   if( ++*number > *allocated ) {
@@ -254,7 +249,7 @@ amiga_asl( char *title, BOOL is_saving ) {
 #else				/* #ifndef __MORPHOS__ */
         AddPart( filename, filereq->fr_File, 1024 );
 #endif				/* #ifndef __MORPHOS__ */
-        widget_filesel_name = strdup( filename );
+        widget_filesel_name = utils_safe_strdup( filename );
 #ifndef __MORPHOS__
         IExec->FreeVec( filename );
 #else				/* #ifndef __MORPHOS__ */
@@ -303,15 +298,10 @@ static int widget_scandir( const char *dir, struct widget_dirent ***namelist,
 #endif				/* #ifdef WIN32 */
 
   while( !done ) {
-    
-#ifndef WIN32
-    char name[ NAME_MAX + 1 ];
-#else				/* #ifndef WIN32 */
-     /* mingw's dirent implementation doesn't have NAME_MAX */
-     char name[ FILENAME_MAX ];
-#endif				/* #ifndef WIN32 */
+    char name[ PATH_MAX ];
 
-    compat_dir_result_t result = compat_readdir( directory, name, sizeof( name ) );
+    compat_dir_result_t result =
+      compat_readdir( directory, name, sizeof( name ) );
 
     switch( result )
     {
@@ -375,7 +365,7 @@ static int widget_scandrives( struct widget_dirent ***namelist )
   unsigned long drivemask;
   int i;
   char drive[3];
-  char *driveletters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
+  const char *driveletters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 
   drive[1] = ':';
   drive[2] = '\0';
@@ -445,7 +435,17 @@ static void widget_scan( char *dir )
 static int
 widget_select_file( const char *name )
 {
-  return( name && strcmp( name, "." ) );
+  if( !name ) return 0;
+
+  /* Skip current directory */
+  if( !strcmp( name, "." ) ) return 0;
+
+#ifndef WIN32
+  /* Skip hidden files/directories */
+  if( strlen( name ) > 1 && name[0] == '.' && name[1] != '.' ) return 0;
+#endif				/* #ifdef WIN32 */
+
+  return 1;
 }
 
 static int widget_scan_compare( const struct widget_dirent **a,
@@ -497,13 +497,13 @@ widget_filesel_draw( void *data )
   /* Create the dialog box */
   error = widget_dialog_with_border( 1, 2, 30, 22 );
   if( error ) {
-    if( directory ) free( directory );
+    free( directory );
     return error; 
   }
 
 #ifdef WIN32
   if( directory == NULL ) {
-    directory = "Drive selection";
+    directory = utils_safe_strdup( "Drive selection" );
   }
 #endif				/* #ifdef WIN32 */
 
@@ -511,6 +511,8 @@ widget_filesel_draw( void *data )
   widget_print_all_filenames( widget_filenames, widget_numfiles,
 			      top_left_file, current_file, directory );
 
+  free( directory );
+
 #endif /* ifndef AMIGA */
 
   return 0;
@@ -606,7 +608,7 @@ static int widget_print_all_filenames( struct widget_dirent **filenames, int n,
 
   /* Print the filenames, mostly normally, but with the currently
      selected file inverted */
-  for( i = top_left; i < n && i < top_left + PAGESIZE; i++ ) {
+  for( i = top_left; i < n && i < top_left + ENTRIES_PER_SCREEN; i++ ) {
     if( i == current ) {
       widget_print_filename( filenames[i], i-top_left, 1 );
     } else {
@@ -807,6 +809,11 @@ http://thread.gmane.org/gmane.comp.gnu.mingw.user/9197
 void
 widget_filesel_keyhandler( input_key key )
 {
+#if !defined AMIGA && !defined __MORPHOS__
+  char *fn, *ptr;
+  char *dirtitle;
+#endif
+
   /* If there are no files (possible on the Wii), can't really do anything */
   if( widget_numfiles == 0 ) {
     if( key == INPUT_KEY_Escape ) widget_end_widget( WIDGET_FINISHED_CANCEL );
@@ -820,8 +827,6 @@ widget_filesel_keyhandler( input_key key )
     widget_end_widget( err );
   }
 #else  /* ifndef AMIGA */
-  char *fn;
-  char *dirtitle;
 
   new_current_file = current_file;
 
@@ -869,13 +874,13 @@ widget_filesel_keyhandler( input_key key )
     break;
 
   case INPUT_KEY_Page_Up:
-    new_current_file = ( current_file > PAGESIZE ) ?
-                       current_file - PAGESIZE     :
+    new_current_file = ( current_file > ENTRIES_PER_SCREEN ) ?
+                       current_file - ENTRIES_PER_SCREEN     :
                        0;
     break;
 
   case INPUT_KEY_Page_Down:
-    new_current_file = current_file + PAGESIZE;
+    new_current_file = current_file + ENTRIES_PER_SCREEN;
     if( new_current_file >= widget_numfiles )
       new_current_file = widget_numfiles - 1;
     break;
@@ -893,17 +898,20 @@ widget_filesel_keyhandler( input_key key )
       widget_text_t text_data;
       text_data.title = title;
       text_data.allow = WIDGET_INPUT_ASCII;
+      text_data.max_length = 30;
       text_data.text[0] = 0;
-      if( widget_do( WIDGET_TYPE_TEXT, &text_data ) ||
+      if( widget_do_text( &text_data ) ||
 	  !widget_text_text || !*widget_text_text      )
 	break;
       if( !compat_is_absolute_path( widget_text_text ) ) {
 							/* relative name */
         /* Get current dir name and allocate space for the leafname */
         fn = widget_getcwd();
+        ptr = fn;
         if( fn )
     	  fn = realloc( fn, strlen( fn ) + strlen( widget_text_text ) + 2 );
         if( !fn ) {
+          free( ptr );
 	  widget_end_widget( WIDGET_FINISHED_CANCEL );
 	  return;
         }
@@ -911,11 +919,7 @@ widget_filesel_keyhandler( input_key key )
         strcat( fn, FUSE_DIR_SEP_STR );
         strcat( fn, widget_text_text );
       } else {						/* absolute name */
-	fn = strdup( widget_text_text );
-        if( !fn ) {
-	  widget_end_widget( WIDGET_FINISHED_CANCEL );
-	  return;
-        }
+	fn = utils_safe_strdup( widget_text_text );
       }
       widget_filesel_name = fn;
       if( exit_all_widgets ) {
@@ -950,7 +954,7 @@ widget_filesel_keyhandler( input_key key )
 
 #ifdef WIN32
   if( is_drivesel ) {
-    dirtitle = "Drive selection";
+    dirtitle = utils_safe_strdup( "Drive selection" );
   } else {
 #endif				/* #ifdef WIN32 */
     dirtitle = widget_getcwd();
@@ -970,9 +974,10 @@ widget_filesel_keyhandler( input_key key )
       widget_print_all_filenames( widget_filenames, widget_numfiles,
 				  top_left_file, new_current_file, dirtitle );
 
-    } else if( new_current_file >= top_left_file+PAGESIZE ) {
+    } else if( new_current_file >= top_left_file+ENTRIES_PER_SCREEN ) {
 
-      top_left_file = new_current_file & ~1; top_left_file -= PAGESIZE - 2;
+      top_left_file = new_current_file & ~1;
+      top_left_file -= ENTRIES_PER_SCREEN - 2;
       widget_print_all_filenames( widget_filenames, widget_numfiles,
 				  top_left_file, new_current_file, dirtitle );
 
@@ -994,5 +999,7 @@ widget_filesel_keyhandler( input_key key )
     current_file = new_current_file;
 
   }
+
+  free( dirtitle );
 #endif /* ifdef AMIGA */
 }
