@@ -1,7 +1,8 @@
 /* debugger.c: the Win32 debugger
-   Copyright (c) 2004 Philip Kendall, Marek Januszewski
-
-   $Id: debugger.c 3920 2008-12-25 23:03:10Z fredm $
+   Copyright (c) 2004-2012 Philip Kendall, Marek Januszewski
+   Copyright (c) 2013-2015 Sergio Baldoví
+   Copyright (c) 2015 Stuart Brady
+   Copyright (c) 2016 BogDan Vatra
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -31,17 +32,18 @@
 */
 
 #include <libspectrum.h>
+#include <stdlib.h>
 #include <tchar.h>
 #include <windows.h>
  
 #include "debugger/debugger.h"
 #include "event.h"
 #include "fuse.h"
-#include "ide/zxcf.h"
-#include "scld.h"
 #include "settings.h"
+#include "peripherals/ide/zxcf.h"
+#include "peripherals/scld.h"
+#include "peripherals/ula.h"
 #include "ui/ui.h"
-#include "ula.h"
 #include "win32internals.h"
 #include "z80/z80.h"
 #include "z80/z80_macros.h"
@@ -78,20 +80,20 @@ static BOOL show_hide_pane( debugger_pane pane, int show );
 static void toggle_display( debugger_pane pane, UINT menu_item_id );
 static int create_register_display( HFONT font );
 /* int create_memory_map( void ); this function is handled by rc */
-static int create_breakpoints();
+static int create_breakpoints( void );
 static int create_disassembly( HFONT font );
 static int create_stack_display( HFONT font );
 static void stack_click( LPNMITEMACTIVATE lpnmitem );
-static int create_events();
+static int create_events( void );
 static void events_click( LPNMITEMACTIVATE lpnmitem );
 /* int create_command_entry( void ); this function is handled by rc */
 /* int create_buttons( void ); this function is handled by rc */
 
 static int activate_debugger( void );
-static int update_memory_map( void );
-static int update_breakpoints();
-static int update_disassembly();
-static int update_events( void );
+static void update_memory_map( void );
+static void update_breakpoints( void );
+static void update_disassembly( void );
+static void update_events( void );
 static void add_event( gpointer data, gpointer user_data GCC_UNUSED );
 static int deactivate_debugger( void );
 
@@ -100,19 +102,26 @@ static int move_disassembly( WPARAM scroll_command );
 static void evaluate_command( void );
 static void win32ui_debugger_done_step( void );
 static void win32ui_debugger_done_continue( void );
-static void win32ui_debugger_break();
+static void win32ui_debugger_break( void );
 static void delete_dialog( void );
 static void win32ui_debugger_done_close( void );
 static INT_PTR CALLBACK win32ui_debugger_proc( HWND hWnd, UINT msg,
                                                WPARAM wParam, LPARAM lParam );
 
+static LRESULT CALLBACK
+disassembly_listview_proc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );
+
 /* The top line of the current disassembly */
 static libspectrum_word disassembly_top;
 
+/* The next line below the current disassembly */
+static libspectrum_word disassembly_bottom;
+
 /* helper constants for disassembly listview's scrollbar */
 static const int disassembly_min = 0x0000;
 static const int disassembly_max = 0xffff;
-static const int disassembly_page = 20;
+/* Visual styles could change visible rows */
+static unsigned int disassembly_page = 20;
 
 /* Have we created the above yet? */
 static int dialog_created = 0;
@@ -154,6 +163,12 @@ ui_debugger_activate( void )
   return 0;
 }
 
+void
+ui_breakpoints_updated( void )
+{
+  /* TODO: Refresh debugger list here */
+}
+
 static int
 hide_hidden_panes( void )
 {
@@ -271,7 +286,7 @@ ui_debugger_deactivate( int interruptable )
 }
 
 static int
-create_dialog()
+create_dialog( void )
 {
   int error;
   debugger_pane i;
@@ -342,7 +357,7 @@ create_register_display( HFONT font )
   /* this display is created in rc, just set the monospaced font */
   size_t i;
 
-  for( i = 0; i < 18; i++ ) {
+  for( i = 0; i < NUM_DBG_REGS; i++ ) {
     win32ui_set_font( fuse_hDBGWnd, IDC_DBG_REG_PC + i, font );
   }
 
@@ -350,12 +365,13 @@ create_register_display( HFONT font )
 }
 
 static int
-create_breakpoints()
+create_breakpoints( void )
 {
   size_t i;
-  
-  TCHAR *breakpoint_titles[] = { "ID", "Type", "Value", "Ignore", "Life",
-                                 "Condition" };
+
+  LPCTSTR breakpoint_titles[] = { _T( "ID" ), _T( "Type" ), _T( "Value" ),
+                                  _T( "Ignore" ), _T( "Life" ), 
+                                  _T( "Condition" ) };
   /* set extended listview style to select full row, when an item is selected */
   DWORD lv_ext_style;
   lv_ext_style = SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_BPS,
@@ -373,7 +389,7 @@ create_breakpoints()
     if( i != 0 )
       lvc.mask |= LVCF_SUBITEM;
     lvc.cx = _tcslen( breakpoint_titles[i] ) * 8 + 10;
-    lvc.pszText = breakpoint_titles[i];
+    lvc.pszText = (LPTSTR)breakpoint_titles[i];
     SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_BPS, LVM_INSERTCOLUMN, i,
                         ( LPARAM ) &lvc );
   }
@@ -386,35 +402,48 @@ create_disassembly( HFONT font )
 {
   size_t i;
 
-  TCHAR *disassembly_titles[] = { TEXT( "Address" ), TEXT( "Instruction" ) };
+  LPCTSTR disassembly_titles[] = { TEXT( "Address" ), TEXT( "Instruction" ) };
 
   /* The disassembly listview itself */
 
+  /* subclass listview to catch keydown and mousewheel messages */
+  HWND hwnd_list = GetDlgItem( fuse_hDBGWnd, IDC_DBG_LV_PC );
+  WNDPROC orig_proc = (WNDPROC) GetWindowLongPtr( hwnd_list, GWLP_WNDPROC );
+  SetProp( hwnd_list, "original_proc", (HANDLE) orig_proc );
+  SetWindowLongPtr( hwnd_list, GWLP_WNDPROC, 
+                    (LONG_PTR) (WNDPROC) disassembly_listview_proc );
+
   /* set extended listview style to select full row, when an item is selected */
   DWORD lv_ext_style;
   lv_ext_style = SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_PC,
                                      LVM_GETEXTENDEDLISTVIEWSTYLE, 0, 0 ); 
   lv_ext_style |= LVS_EX_FULLROWSELECT;
+  lv_ext_style |= LVS_EX_DOUBLEBUFFER;
   SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_PC,
                       LVM_SETEXTENDEDLISTVIEWSTYLE, 0, lv_ext_style ); 
 
+  win32ui_set_font( fuse_hDBGWnd, IDC_DBG_LV_PC, font );
+
   /* create columns */
   LVCOLUMN lvc;
-  lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT ;
+  lvc.mask = LVCF_FMT | LVCF_TEXT;
   lvc.fmt = LVCFMT_LEFT;
 
   for( i = 0; i < 2; i++ ) {
-    if( i != 0 ) {
-      lvc.mask |= LVCF_SUBITEM;
-    }
-    lvc.cx = ( i == 0 ) ? 65 : 125;
-    lvc.pszText = disassembly_titles[i];
+    if( i != 0 ) lvc.mask |= LVCF_SUBITEM;
+    lvc.pszText = (LPTSTR)disassembly_titles[i];
     SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_PC, LVM_INSERTCOLUMN, i,
                         ( LPARAM ) &lvc );
   }
-  
-  win32ui_set_font( fuse_hDBGWnd, IDC_DBG_LV_PC, font );
-  
+
+  /* Set columns width */
+  ListView_SetColumnWidth( hwnd_list, 0, LVSCW_AUTOSIZE_USEHEADER );
+  ListView_SetColumnWidth( hwnd_list, 1, LVSCW_AUTOSIZE_USEHEADER );
+
+  /* Recalculate visible rows, Visual Styles could change rows height */
+  disassembly_page = SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_PC,
+                                         LVM_GETCOUNTPERPAGE, 0, 0 );
+
   /* The disassembly scrollbar */
   SCROLLINFO si;
   si.cbSize = sizeof(si); 
@@ -433,9 +462,9 @@ static int
 create_stack_display( HFONT font )
 {
   size_t i;
-  
-  TCHAR *stack_titles[] = { "Address", "Value" };
-  
+
+  LPCTSTR stack_titles[] = { _T( "Address" ), _T( "Value" ) };
+
   /* set extended listview style to select full row, when an item is selected */
   DWORD lv_ext_style;
   lv_ext_style = SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_STACK,
@@ -444,22 +473,24 @@ create_stack_display( HFONT font )
   SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_STACK,
                       LVM_SETEXTENDEDLISTVIEWSTYLE, 0, lv_ext_style ); 
 
+  win32ui_set_font( fuse_hDBGWnd, IDC_DBG_LV_STACK, font );
+
   /* create columns */
   LVCOLUMN lvc;
-  lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT ;
+  lvc.mask = LVCF_FMT | LVCF_TEXT;
   lvc.fmt = LVCFMT_LEFT;
 
   for( i = 0; i < 2; i++ ) {
-    if( i != 0 )
-      lvc.mask |= LVCF_SUBITEM;
-    lvc.cx = ( i == 0 ) ? 65 : 55;
-
-    lvc.pszText = stack_titles[i];
+    if( i != 0 ) lvc.mask |= LVCF_SUBITEM;
+    lvc.pszText = (LPTSTR)stack_titles[i];
     SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_STACK, LVM_INSERTCOLUMN, i,
                         ( LPARAM ) &lvc );
   }
-  
-  win32ui_set_font( fuse_hDBGWnd, IDC_DBG_LV_STACK, font );
+
+  /* Set columns width */
+  HWND hwnd_list = GetDlgItem( fuse_hDBGWnd, IDC_DBG_LV_STACK );
+  ListView_SetColumnWidth( hwnd_list, 0, LVSCW_AUTOSIZE_USEHEADER );
+  ListView_SetColumnWidth( hwnd_list, 1, LVSCW_AUTOSIZE_USEHEADER );
 
   return 0;
 }
@@ -467,19 +498,29 @@ create_stack_display( HFONT font )
 static void
 stack_click( LPNMITEMACTIVATE lpnmitem )
 {
-  libspectrum_word address, destination;
-  int error, row;
+  libspectrum_word destination;
+  int retval, error, row;
+  char *stopstring;
+  TCHAR buffer[255];
+  LVITEM li;
 
   row = lpnmitem->iItem;
   if( row < 0 ) return;
 
-  address = SP + ( 19 - row ) * 2;
+  li.iSubItem = 1;
+  li.pszText = buffer;
+  li.cchTextMax = 255;
+  retval = SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_STACK,
+                                 LVM_GETITEMTEXT, row, ( LPARAM ) &li );
+  if( !retval ) {
+    ui_error( UI_ERROR_ERROR, "couldn't get text for row %d", row );
+    return;
+  }
 
-  destination =
-    readbyte_internal( address ) + readbyte_internal( address + 1 ) * 0x100;
+  destination = strtoul( buffer, &stopstring, 16 );
 
   error = debugger_breakpoint_add_address(
-    DEBUGGER_BREAKPOINT_TYPE_EXECUTE, -1, destination, 0,
+    DEBUGGER_BREAKPOINT_TYPE_EXECUTE, memory_source_any, 0, destination, 0,
     DEBUGGER_BREAKPOINT_LIFE_ONESHOT, NULL
   );
   if( error ) return;
@@ -488,12 +529,11 @@ stack_click( LPNMITEMACTIVATE lpnmitem )
 }
 
 static int
-create_events()
+create_events( void )
 {
   size_t i;
-  
-  TCHAR *titles[] = { "Time", "Type" };
-  
+  LPCTSTR titles[] = { _T( "Time" ), _T( "Type" ) };
+
   /* set extended listview style to select full row, when an item is selected */
   DWORD lv_ext_style;
   lv_ext_style = SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_EVENTS,
@@ -504,18 +544,21 @@ create_events()
 
   /* create columns */
   LVCOLUMN lvc;
-  lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT ;
+  lvc.mask = LVCF_FMT | LVCF_TEXT;
   lvc.fmt = LVCFMT_LEFT;
 
   for( i = 0; i < 2; i++ ) {
-    if( i != 0 )
-      lvc.mask |= LVCF_SUBITEM;
-    lvc.cx = ( i == 0 )? 40 : 75;
-    lvc.pszText = titles[i];
+    if( i != 0 ) lvc.mask |= LVCF_SUBITEM;
+    lvc.pszText = (LPTSTR)titles[i];
     SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_EVENTS, LVM_INSERTCOLUMN, i,
                         ( LPARAM ) &lvc );
   }
-  
+
+  /* Set columns width */
+  HWND hwnd_list = GetDlgItem( fuse_hDBGWnd, IDC_DBG_LV_EVENTS );
+  ListView_SetColumnWidth( hwnd_list, 0, LVSCW_AUTOSIZE_USEHEADER );
+  ListView_SetColumnWidth( hwnd_list, 1, LVSCW_AUTOSIZE_USEHEADER );
+
   return 0;
 }
 
@@ -571,7 +614,6 @@ ui_debugger_update( void )
   TCHAR *disassembly_text[2] = { &buffer[0], &buffer[40] };
   libspectrum_word address;
   int capabilities; size_t length;
-  int error;
 
   const char *register_name[] = { TEXT( "PC" ), TEXT( "SP" ),
 				  TEXT( "AF" ), TEXT( "AF'" ),
@@ -606,7 +648,9 @@ ui_debugger_update( void )
   SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_REG_R,
                       WM_SETTEXT, (WPARAM) 0, (LPARAM) buffer );
 
-  /* FIXME: doesnt' look like T-states fits in with monospaced font? */
+  _sntprintf( buffer, 80, TEXT( "Halted %d" ), z80.halted );
+  SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_REG_HALTED,
+                      WM_SETTEXT, (WPARAM) 0, (LPARAM) buffer );
   _sntprintf( buffer, 80, TEXT( "T-states %5d" ), tstates );
   SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_REG_T_STATES,
                       WM_SETTEXT, (WPARAM) 0, (LPARAM) buffer );
@@ -673,13 +717,9 @@ ui_debugger_update( void )
   SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_REG_ULA,
                       WM_SETTEXT, (WPARAM) 0, (LPARAM) buffer );
 
-  /* Update the memory map display */
-  error = update_memory_map(); if( error ) return error;
-
-  error = update_breakpoints(); if( error ) return error;
-
-  /* Update the disassembly */
-  error = update_disassembly(); if( error ) return error;
+  update_memory_map();
+  update_breakpoints();
+  update_disassembly();
 
   /* And the stack display */
   SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_STACK,
@@ -709,52 +749,85 @@ ui_debugger_update( void )
   }
 
   /* And the events display */
-  error = update_events(); if( error ) return error;
+  update_events();
 
   return 0;
 }
 
-static int
+static void
 update_memory_map( void )
 {
-  size_t i;
   TCHAR buffer[ 40 ];
+  int source, page_num, writable, contended;
+  libspectrum_word offset;
+  size_t i, j, block, row;
 
-  for( i = 0; i < 8; i++ ) {
+  source = page_num = writable = contended = -1;
+  offset = 0;
+  row = 0;
 
-    _sntprintf( buffer, 40, format_16_bit(), (unsigned)i * 0x2000 );
-    SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( i * 4 ), 
-                        WM_SETTEXT, ( WPARAM ) 0, ( LPARAM ) buffer );
+  for( block = 0; block < MEMORY_PAGES_IN_64K && row < 8; block++ ) {
+    memory_page *page = &memory_map_read[block];
 
-    /* FIXME: memory_bank_name is not unicode */
-    _sntprintf( buffer, 40, TEXT( "%s %d" ), memory_bank_name( &memory_map_read[i] ),
-	        memory_map_read[i].page_num );
-    SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( i * 4 ) + 1, 
-                        WM_SETTEXT, ( WPARAM ) 0, ( LPARAM ) buffer );
+    if( page->source != source ||
+      page->page_num != page_num ||
+      page->offset != offset ||
+      page->writable != writable ||
+      page->contended != contended ) {
 
-    SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( i * 4 ) + 2, 
-                        WM_SETTEXT, (WPARAM) 0,
-                        ( LPARAM ) ( memory_map_read[i].writable
-                        ? TEXT( "Y" ) : TEXT( "N" ) ) );
+      _sntprintf( buffer, 40, format_16_bit(),
+                  (unsigned)block * MEMORY_PAGE_SIZE );
+      SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( row * 4 ), 
+                          WM_SETTEXT, ( WPARAM ) 0, ( LPARAM ) buffer );
 
-    SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( i * 4 ) + 2, 
-                        WM_SETTEXT, (WPARAM) 0,
-                        ( LPARAM ) ( memory_map_read[i].contended
-                        ? TEXT( "Y" ) : TEXT( "N" ) ) );
+      /* FIXME: memory_source_description is not unicode */
+      _snprintf( buffer, 40, TEXT( "%s %d" ),
+                 memory_source_description( page->source ), page->page_num );
+
+      SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( row * 4 ) + 1,
+                          WM_SETTEXT, ( WPARAM ) 0, ( LPARAM ) buffer );
+
+      SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( row * 4 ) + 2,
+                          WM_SETTEXT, (WPARAM) 0,
+                          ( LPARAM ) ( page->writable
+                          ? TEXT( "Y" ) : TEXT( "N" ) ) );
+
+      SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( row * 4 ) + 3,
+                          WM_SETTEXT, (WPARAM) 0,
+                          ( LPARAM ) ( page->contended
+                          ? TEXT( "Y" ) : TEXT( "N" ) ) );
+      row++;
+
+      source = page->source;
+      page_num = page->page_num;
+      writable = page->writable;
+      contended = page->contended;
+      offset = page->offset;
+    }
+
+    /* We expect the next page to have an increased offset */
+    offset += MEMORY_PAGE_SIZE;
+  }
+
+  /* Hide unused rows */
+  for( i = row; i < 8; i++ ) {
+    for( j = 0; j < 4; j++ ) {
+      SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_MAP11 + ( i * 4 ) + j,
+                          WM_SETTEXT, (WPARAM) 0, (LPARAM) NULL );
+    }
   }
 
-  return 0;
 }
 
-static int
-update_breakpoints()
+static void
+update_breakpoints( void )
 {
   /* FIXME: review this function for unicode compatibility */
   TCHAR buffer[ 1024 ],
     *breakpoint_text[6] = { &buffer[  0], &buffer[ 40], &buffer[80],
 			    &buffer[120], &buffer[160], &buffer[200] };
   GSList *ptr;
-  TCHAR format_string[ 1024 ], page[ 1024 ];
+  TCHAR format_string[ 1024 ];
 
   LV_ITEM lvi;
   lvi.mask = LVIF_TEXT;
@@ -777,14 +850,15 @@ update_breakpoints()
     case DEBUGGER_BREAKPOINT_TYPE_EXECUTE:
     case DEBUGGER_BREAKPOINT_TYPE_READ:
     case DEBUGGER_BREAKPOINT_TYPE_WRITE:
-      if( bp->value.address.page == -1 ) {
-	_sntprintf( breakpoint_text[2], 40, format_16_bit(),
-		    bp->value.address.offset );
+      if( bp->value.address.source == memory_source_any ) {
+        _sntprintf( breakpoint_text[2], 40, format_16_bit(),
+        bp->value.address.offset );
       } else {
-	debugger_breakpoint_decode_page( page, 1024, bp->value.address.page );
-	_sntprintf( format_string, 1024, "%%s:%s", format_16_bit() );
-	_sntprintf( breakpoint_text[2], 40, format_string, page,
-		    bp->value.address.offset );
+        snprintf( format_string, 1024, "%%s:%s:%s",
+                  format_16_bit(), format_16_bit() );
+        snprintf( breakpoint_text[2], 40, format_string,
+                  memory_source_description( bp->value.address.source ),
+                  bp->value.address.page, bp->value.address.offset );
       }
       break;
 
@@ -830,12 +904,10 @@ update_breakpoints()
                             ( LPARAM ) &lvi );
     }
   }
-
-  return 0;
 }
 
-static int
-update_disassembly()
+static void
+update_disassembly( void )
 {
   size_t i, length; libspectrum_word address;
   TCHAR buffer[80];
@@ -847,7 +919,7 @@ update_disassembly()
   LV_ITEM lvi;
   lvi.mask = LVIF_TEXT;
 
-  for( i = 0, address = disassembly_top; i < 20; i++ ) {
+  for( i = 0, address = disassembly_top; i < disassembly_page; i++ ) {
     int l;
     _sntprintf( disassembly_text[0], 40, format_16_bit(), address );
     debugger_disassemble( disassembly_text[1], 40, &length, address );
@@ -860,8 +932,7 @@ update_disassembly()
     address += length;
 
     /* append the item */
-    lvi.iItem = SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_PC,
-                                    LVM_GETITEMCOUNT, 0, 0 );
+    lvi.iItem = i;
     lvi.iSubItem = 0;
     lvi.pszText = disassembly_text[0];
     SendDlgItemMessage( fuse_hDBGWnd, IDC_DBG_LV_PC, LVM_INSERTITEM, 0,
@@ -872,10 +943,10 @@ update_disassembly()
                         ( LPARAM ) &lvi );
   }
 
-  return 0;
+  disassembly_bottom = address;
 }
 
-static int
+static void
 update_events( void )
 {
   /* clear the listview */
@@ -883,8 +954,6 @@ update_events( void )
                       LVM_DELETEALLITEMS, 0, 0 );
 
   event_foreach( add_event, NULL );
-
-  return 0;
 }
 
 static void
@@ -931,12 +1000,18 @@ deactivate_debugger( void )
 int
 ui_debugger_disassemble( libspectrum_word address )
 {
+  /* Note: the scroll bar can not cope with "upper bound - page_size" value and
+     higher. PC register, key and wheel scrolling are fine with that */
   SCROLLINFO si;
   si.cbSize = sizeof(si); 
   si.fMask = SIF_POS; 
   si.nPos = disassembly_top = address;
   SetScrollInfo( GetDlgItem( fuse_hDBGWnd, IDC_DBG_SB_PC ),
                  SB_CTL, &si, TRUE );
+
+  /* And update the disassembly if the debugger is active */
+  if( debugger_active ) update_disassembly();
+
   return 0;
 }
 
@@ -944,80 +1019,64 @@ ui_debugger_disassemble( libspectrum_word address )
 static int
 move_disassembly( WPARAM scroll_command )
 {
-  SCROLLINFO si;
-  si.cbSize = sizeof(si); 
-  si.fMask = SIF_POS; 
-  GetScrollInfo( GetDlgItem( fuse_hDBGWnd, IDC_DBG_SB_PC ), SB_CTL, &si );
+  libspectrum_word address;
+  int cursor_row;
 
-  int value = si.nPos;
-  
   /* in Windows we have to read the command and scroll the scrollbar manually */
   switch( LOWORD( scroll_command ) ) {
     case SB_BOTTOM:
-      value = disassembly_max;
+      if( disassembly_bottom == disassembly_min ) return 0;
+      address = debugger_search_instruction( disassembly_min,
+                                             -disassembly_page );
       break;
     case SB_TOP:
-      value = disassembly_min;
+      if( disassembly_top == disassembly_min ) return 0;
+      address = disassembly_min;
       break;
     case SB_LINEDOWN:
-      value++;
+      if( disassembly_bottom == disassembly_min ) return 0;
+      address = debugger_search_instruction( disassembly_top, 1 );
       break;
     case SB_LINEUP:
-      value--;
+      if( disassembly_top == disassembly_min ) return 0;
+      address = debugger_search_instruction( disassembly_top, -1);
       break;
     case SB_PAGEUP:
-      value -= disassembly_page;
+      address = debugger_search_instruction( disassembly_top,
+                                             -disassembly_page );
       break;
     case SB_PAGEDOWN:
-      value += disassembly_page;
+      address = debugger_search_instruction( disassembly_top,
+                                             disassembly_page );
       break;
     case SB_THUMBPOSITION:
     case SB_THUMBTRACK:
-      value = HIWORD( scroll_command );
+      /* just set disassembly_top to that value */
+      address = HIWORD( scroll_command );
+
+      /* The scrollbar should constrain to min/max values */
+      if( address > disassembly_max - disassembly_page )
+        address = debugger_search_instruction( disassembly_min,
+                                               -disassembly_page );
       break;
     default:
       return 1;
   }
-  if( value > disassembly_max ) value = disassembly_max;
-  if( value < disassembly_min ) value = disassembly_min;
-  /* FIXME when holding sb's down arrow at 0xffff, 0x0000 is showing */
 
-  size_t length;
+  /* Get selected row */
+  cursor_row = ListView_GetNextItem( GetDlgItem( fuse_hDBGWnd, IDC_DBG_LV_PC ),
+                                     -1, LVNI_SELECTED );
 
-  /* disassembly_top < value < disassembly_top + 1 => 'down' button pressed
-     Move the disassembly on by one instruction */
-  if( value > disassembly_top && value - disassembly_top < 1 ) {
-
-    debugger_disassemble( NULL, 0, &length, disassembly_top );
-    ui_debugger_disassemble( disassembly_top + length );
-
-  /* disassembly_top - 1 < value < disassembly_top => 'up' button pressed
-     
-     See notes regarding how this function works in GTK's move_disassembly
-  */
-  } else if( value < disassembly_top && disassembly_top - value < 1 ) {
-
-    size_t i, longest = 1;
-
-    for( i = 1; i <= 8; i++ ) {
-
-      debugger_disassemble( NULL, 0, &length, disassembly_top - i );
-      if( length == i ) longest = i;
-
-    }
-
-    ui_debugger_disassemble( disassembly_top - longest );
-
-  /* Anything else, just set disassembly_top to that value */
-  } else {
-
-    ui_debugger_disassemble( value );
+  /* Scroll to new position */
+  ui_debugger_disassemble( address );
 
+  /* Mark selected row */
+  if( cursor_row >= 0 ) {
+    ListView_SetItemState( GetDlgItem( fuse_hDBGWnd, IDC_DBG_LV_PC ),
+                           cursor_row, LVIS_FOCUSED|LVIS_SELECTED,
+                           LVIS_FOCUSED|LVIS_SELECTED );
   }
 
-  /* And update the disassembly if the debugger is active */
-  if( debugger_active ) update_disassembly();
-        
   return 0;
 }
 
@@ -1087,13 +1146,14 @@ win32ui_debugger_done_close( void )
 }
 
 static INT_PTR CALLBACK
-win32ui_debugger_proc( HWND hWnd, UINT msg,
+win32ui_debugger_proc( HWND hWnd GCC_UNUSED, UINT msg,
                        WPARAM wParam, LPARAM lParam )
 {
   switch( msg ) {
     case WM_COMMAND:
       switch( LOWORD( wParam ) ) {
         case IDCLOSE:
+        case IDCANCEL:
           win32ui_debugger_done_close();
           return 0;
         case IDC_DBG_BTN_CONT:
@@ -1159,3 +1219,115 @@ win32ui_debugger_proc( HWND hWnd, UINT msg,
   }
   return FALSE;
 }
+
+static LRESULT CALLBACK
+disassembly_key_press( HWND hWnd, WPARAM wParam )
+{
+  libspectrum_word initial_top, address;
+  int cursor_row;
+
+  initial_top = disassembly_top;
+
+  /* Get selected row */
+  cursor_row = ListView_GetNextItem( hWnd, -1, LVNI_SELECTED );
+
+  switch( wParam ) {
+  case VK_DOWN:
+    if( cursor_row == disassembly_page - 1 ) {
+      address = debugger_search_instruction( disassembly_top, 1 );
+      ui_debugger_disassemble( address );
+    }
+    break;
+
+  case VK_UP:
+    if( cursor_row == 0 ) {
+      address = debugger_search_instruction( disassembly_top, -1 );
+      ui_debugger_disassemble( address );
+    }
+    break;
+
+  case VK_NEXT:
+    ui_debugger_disassemble( disassembly_bottom );
+    break;
+
+  case VK_PRIOR:
+    address = debugger_search_instruction( disassembly_top,
+                                           -disassembly_page );
+    ui_debugger_disassemble( address );
+    break;
+
+  case VK_HOME:
+    cursor_row = 0;
+    ui_debugger_disassemble( disassembly_min );
+    break;
+
+  case VK_END:
+    cursor_row = disassembly_page - 1;
+    address = debugger_search_instruction( disassembly_min,
+                                           -disassembly_page );
+    ui_debugger_disassemble( address );
+    break;
+  }
+
+  if( initial_top != disassembly_top ) {
+    /* Mark selected row */
+    if( cursor_row >= 0 ) {
+      ListView_SetItemState( hWnd, cursor_row, LVIS_FOCUSED|LVIS_SELECTED,
+                             LVIS_FOCUSED|LVIS_SELECTED );
+    }
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+static LRESULT CALLBACK
+disassembly_wheel_scroll( HWND hWnd, WPARAM wParam )
+{
+  libspectrum_word address;
+  int cursor_row;
+  short delta;
+
+  /* Get selected row */
+  cursor_row = ListView_GetNextItem( hWnd, -1, LVNI_SELECTED );
+
+  /* Convert wheel displacement to instruction displacement */
+  delta = (short) HIWORD( wParam ) / WHEEL_DELTA;
+
+  /* Scroll to new position */
+  address = debugger_search_instruction( disassembly_top, -delta );
+  ui_debugger_disassemble( address );
+
+  /* Mark selected row */
+  if( cursor_row >= 0 ) {
+    ListView_SetItemState( hWnd, cursor_row, LVIS_FOCUSED|LVIS_SELECTED,
+                           LVIS_FOCUSED|LVIS_SELECTED );
+  }
+
+  return TRUE;
+}
+
+static LRESULT CALLBACK
+disassembly_listview_proc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
+{
+  WNDPROC orig_proc;
+
+  switch( msg ) {
+  case WM_DESTROY:
+    orig_proc = (WNDPROC) GetProp( hWnd, "original_proc" );
+    SetWindowLongPtr( hWnd, GWLP_WNDPROC, (LONG_PTR) orig_proc );
+    RemoveProp( hWnd, "original_proc" );
+    break;
+
+  case WM_KEYDOWN:
+    if( disassembly_key_press( hWnd, wParam ) ) return 0;
+    break;
+
+  case WM_MOUSEWHEEL:
+    disassembly_wheel_scroll( hWnd, wParam );
+    return 0;
+  }
+
+  orig_proc = (WNDPROC) GetProp( hWnd, "original_proc" );
+  return CallWindowProc( orig_proc, hWnd, msg, wParam, lParam );
+}
