@@ -1,9 +1,8 @@
 #!/usr/bin/perl -w
 
 # z80.pl: generate C code for Z80 opcodes
-# $Id: z80.pl,v 1.64 2004/07/07 09:27:58 pak21 Exp $
 
-# Copyright (c) 1999-2003 Philip Kendall
+# Copyright (c) 1999-2015 Philip Kendall
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -15,19 +14,16 @@
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
 
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 49 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+# You should have received a copy of the GNU General Public License along
+# with this program; if not, write to the Free Software Foundation, Inc.,
+# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
 # Author contact information:
 
-# E-mail: pak21-fuse@srcf.ucam.org
-# Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+# E-mail: philip-fuse@shadowmagic.org.uk
 
 use strict;
 
-use lib '../perl';
-
 use Fuse;
 
 # The status of which flags relates to which condition
@@ -61,8 +57,9 @@ sub arithmetic_logical ($$$) {
       {
 	libspectrum_byte offset, bytetemp;
 	offset = readbyte( PC );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); contend_read( PC, 1 );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); PC++;
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 );
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 );
+	contend_read_no_mreq( PC, 1 ); PC++;
 	bytetemp = readbyte( REGISTER + (libspectrum_signed_byte)offset );
 	$opcode(bytetemp);
       }
@@ -78,9 +75,27 @@ CODE
 CODE
 	}
     } elsif( $opcode eq 'ADD' ) {
-	print "      ${opcode}16($arg1,$arg2);\n";
+	print << "CODE";
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      ${opcode}16($arg1,$arg2);
+CODE
     } elsif( $arg1 eq 'HL' and length $arg2 == 2 ) {
-	print "      tstates += 7;\n      ${opcode}16($arg2);\n";
+	print << "CODE";
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      ${opcode}16($arg2);
+CODE
     }
 }
 
@@ -119,8 +134,9 @@ sub cpi_cpd ($) {
 	  lookup = ( (        A & 0x08 ) >> 3 ) |
 	           ( (  (value) & 0x08 ) >> 2 ) |
 	           ( ( bytetemp & 0x08 ) >> 1 );
-	contend_read( HL, 1 ); contend_read( HL, 1 ); contend_read( HL, 1 );
-	contend_read( HL, 1 ); contend_read( HL, 1 );
+	contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
+	contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
+	contend_read_no_mreq( HL, 1 );
 	HL$modifier; BC--;
 	F = ( F & FLAG_C ) | ( BC ? ( FLAG_V | FLAG_N ) : FLAG_N ) |
 	  halfcarry_sub_table[lookup] | ( bytetemp ? 0 : FLAG_Z ) |
@@ -143,19 +159,22 @@ sub cpir_cpdr ($) {
 	  lookup = ( (        A & 0x08 ) >> 3 ) |
 		   ( (  (value) & 0x08 ) >> 2 ) |
 		   ( ( bytetemp & 0x08 ) >> 1 );
-	contend_read( HL, 1 ); contend_read( HL, 1 ); contend_read( HL, 1 );
-	contend_read( HL, 1 ); contend_read( HL, 1 );
-	HL$modifier; BC--;
+	contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
+	contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
+	contend_read_no_mreq( HL, 1 );
+	BC--;
 	F = ( F & FLAG_C ) | ( BC ? ( FLAG_V | FLAG_N ) : FLAG_N ) |
 	  halfcarry_sub_table[lookup] | ( bytetemp ? 0 : FLAG_Z ) |
 	  ( bytetemp & FLAG_S );
 	if(F & FLAG_H) bytetemp--;
 	F |= ( bytetemp & FLAG_3 ) | ( (bytetemp&0x02) ? FLAG_5 : 0 );
 	if( ( F & ( FLAG_V | FLAG_Z ) ) == FLAG_V ) {
-	  contend_read( HL, 1 ); contend_read( HL, 1 ); contend_read( HL, 1 );
-	  contend_read( HL, 1 ); contend_read( HL, 1 );
+	  contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
+	  contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
+	  contend_read_no_mreq( HL, 1 );
 	  PC-=2;
 	}
+	HL$modifier;
       }
 CODE
 }
@@ -169,12 +188,16 @@ sub inc_dec ($$) {
     if( length $arg == 1 or $arg =~ /^REGISTER[HL]$/ ) {
 	print "      $opcode($arg);\n";
     } elsif( length $arg == 2 or $arg eq 'REGISTER' ) {
-	print "      tstates += 2;\n      ${arg}$modifier;\n";
+	print << "CODE";
+	contend_read_no_mreq( IR, 1 );
+	contend_read_no_mreq( IR, 1 );
+	${arg}$modifier;
+CODE
     } elsif( $arg eq '(HL)' ) {
 	print << "CODE";
       {
 	libspectrum_byte bytetemp = readbyte( HL );
-	contend_read( HL, 1 );
+	contend_read_no_mreq( HL, 1 );
 	$opcode(bytetemp);
 	writebyte(HL,bytetemp);
       }
@@ -185,11 +208,12 @@ CODE
 	libspectrum_byte offset, bytetemp;
 	libspectrum_word wordtemp;
 	offset = readbyte( PC );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); contend_read( PC, 1 );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); PC++;
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 );
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 );
+	contend_read_no_mreq( PC, 1 ); PC++;
 	wordtemp = REGISTER + (libspectrum_signed_byte)offset;
 	bytetemp = readbyte( wordtemp );
-	contend_read( wordtemp, 1 );
+	contend_read_no_mreq( wordtemp, 1 );
 	$opcode(bytetemp);
 	writebyte(wordtemp,bytetemp);
       }
@@ -202,20 +226,22 @@ sub ini_ind ($) {
 
     my( $opcode ) = @_;
 
-    my $modifier = ( $opcode eq 'INI' ? '++' : '--' );
+    my $modifier = ( $opcode eq 'INI' ? '+' : '-' );
 
     print << "CODE";
       {
-	libspectrum_word initemp;
+	libspectrum_byte initemp, initemp2;
 
-	tstates++;
-	ula_contend_port( BC );
+	contend_read_no_mreq( IR, 1 );
 	initemp = readport( BC );
 	writebyte( HL, initemp );
 
-	B--; HL$modifier;
-	F = ( initemp & 0x80 ? FLAG_N : 0 ) | sz53_table[B];
-	/* C,H and P/V flags not implemented */
+        B--; HL$modifier$modifier;
+        initemp2 = initemp + C $modifier 1;
+	F = ( initemp & 0x80 ? FLAG_N : 0 ) |
+            ( ( initemp2 < initemp ) ? FLAG_H | FLAG_C : 0 ) |
+            ( parity_table[ ( initemp2 & 0x07 ) ^ B ] ? FLAG_P : 0 ) |
+            sz53_table[B];
       }
 CODE
 }
@@ -224,27 +250,30 @@ sub inir_indr ($) {
 
     my( $opcode ) = @_;
 
-    my $modifier = ( $opcode eq 'INIR' ? '++' : '--' );
+    my $modifier = ( $opcode eq 'INIR' ? '+' : '-' );
 
     print << "CODE";
       {
-	libspectrum_word initemp;
+	libspectrum_byte initemp, initemp2;
 
-	tstates++;
-	ula_contend_port( BC );
+	contend_read_no_mreq( IR, 1 );
 	initemp = readport( BC );
 	writebyte( HL, initemp );
 
-	B--; HL$modifier;
-	F = ( initemp & 0x80 ? FLAG_N : 0 ) | sz53_table[B];
-	/* C,H and P/V flags not implemented */
+	B--;
+        initemp2 = initemp + C $modifier 1;
+	F = ( initemp & 0x80 ? FLAG_N : 0 ) |
+            ( ( initemp2 < initemp ) ? FLAG_H | FLAG_C : 0 ) |
+            ( parity_table[ ( initemp2 & 0x07 ) ^ B ] ? FLAG_P : 0 ) |
+            sz53_table[B];
 
 	if( B ) {
-	  contend_write( HL, 1 ); contend_write( HL, 1 );
-	  contend_write( HL, 1 ); contend_write( HL, 1 );
-	  contend_write( HL, 1 );
+	  contend_write_no_mreq( HL, 1 ); contend_write_no_mreq( HL, 1 );
+	  contend_write_no_mreq( HL, 1 ); contend_write_no_mreq( HL, 1 );
+	  contend_write_no_mreq( HL, 1 );
 	  PC -= 2;
 	}
+        HL$modifier$modifier;
       }
 CODE
 }
@@ -261,7 +290,7 @@ sub ldi_ldd ($) {
 	libspectrum_byte bytetemp=readbyte( HL );
 	BC--;
 	writebyte(DE,bytetemp);
-	contend_write( DE, 1 ); contend_write( DE, 1 );
+	contend_write_no_mreq( DE, 1 ); contend_write_no_mreq( DE, 1 );
 	DE$modifier; HL$modifier;
 	bytetemp += A;
 	F = ( F & ( FLAG_C | FLAG_Z | FLAG_S ) ) | ( BC ? FLAG_V : 0 ) |
@@ -280,17 +309,18 @@ sub ldir_lddr ($) {
       {
 	libspectrum_byte bytetemp=readbyte( HL );
 	writebyte(DE,bytetemp);
-	contend_write( DE, 1 ); contend_write( DE, 1 );
-	HL$modifier; DE$modifier; BC--;
+	contend_write_no_mreq( DE, 1 ); contend_write_no_mreq( DE, 1 );
+	BC--;
 	bytetemp += A;
 	F = ( F & ( FLAG_C | FLAG_Z | FLAG_S ) ) | ( BC ? FLAG_V : 0 ) |
 	  ( bytetemp & FLAG_3 ) | ( (bytetemp & 0x02) ? FLAG_5 : 0 );
 	if(BC) {
-	  contend_write( DE, 1 ); contend_write( DE, 1 );
-	  contend_write( DE, 1 ); contend_write( DE, 1 );
-	  contend_write( DE, 1 );
+	  contend_write_no_mreq( DE, 1 ); contend_write_no_mreq( DE, 1 );
+	  contend_write_no_mreq( DE, 1 ); contend_write_no_mreq( DE, 1 );
+	  contend_write_no_mreq( DE, 1 );
 	  PC-=2;
 	}
+        HL$modifier; DE$modifier;
       }
 CODE
 }
@@ -303,21 +333,24 @@ sub otir_otdr ($) {
 
     print << "CODE";
       {
-	libspectrum_word outitemp;
+	libspectrum_byte outitemp, outitemp2;
 
-	tstates++;
+	contend_read_no_mreq( IR, 1 );
 	outitemp = readbyte( HL );
 	B--;	/* This does happen first, despite what the specs say */
-	ula_contend_port( BC );
 	writeport(BC,outitemp);
 
 	HL$modifier;
-	F = (outitemp & 0x80 ? FLAG_N : 0 ) | sz53_table[B];
-	/* C,H and P/V flags not implemented */
+        outitemp2 = outitemp + L;
+	F = ( outitemp & 0x80 ? FLAG_N : 0 ) |
+            ( ( outitemp2 < outitemp ) ? FLAG_H | FLAG_C : 0 ) |
+            ( parity_table[ ( outitemp2 & 0x07 ) ^ B ] ? FLAG_P : 0 ) |
+            sz53_table[B];
 
 	if( B ) {
-	  contend_read( HL, 1 ); contend_read( HL, 1 ); contend_read( HL, 1 );
-	  contend_read( HL, 1 ); contend_read( HL, 1 );
+	  contend_read_no_mreq( BC, 1 ); contend_read_no_mreq( BC, 1 );
+	  contend_read_no_mreq( BC, 1 ); contend_read_no_mreq( BC, 1 );
+	  contend_read_no_mreq( BC, 1 );
 	  PC -= 2;
 	}
       }
@@ -332,17 +365,19 @@ sub outi_outd ($) {
 
     print << "CODE";
       {
-	libspectrum_word outitemp;
+	libspectrum_byte outitemp, outitemp2;
 
-	tstates++;
+	contend_read_no_mreq( IR, 1 );
 	outitemp = readbyte( HL );
 	B--;	/* This does happen first, despite what the specs say */
-	ula_contend_port( BC );
 	writeport(BC,outitemp);
 
 	HL$modifier;
-	F = (outitemp & 0x80 ? FLAG_N : 0 ) | sz53_table[B];
-	/* C,H and P/V flags not implemented */
+        outitemp2 = outitemp + L;
+	F = ( outitemp & 0x80 ? FLAG_N : 0 ) |
+            ( ( outitemp2 < outitemp ) ? FLAG_H | FLAG_C : 0 ) |
+            ( parity_table[ ( outitemp2 & 0x07 ) ^ B ] ? FLAG_P : 0 ) |
+            sz53_table[B];
       }
 CODE
 }
@@ -386,7 +421,7 @@ sub res_set ($$$) {
 	print << "CODE";
       {
 	libspectrum_byte bytetemp = readbyte( HL );
-	contend_read( HL, 1 );
+	contend_read_no_mreq( HL, 1 );
 	writebyte( HL, bytetemp $operator $hex_mask );
       }
 CODE
@@ -395,7 +430,7 @@ CODE
       {
 	libspectrum_byte bytetemp;
 	bytetemp = readbyte( tempaddr );
-	contend_read( tempaddr, 1 );
+	contend_read_no_mreq( tempaddr, 1 );
 	writebyte( tempaddr, bytetemp $operator $hex_mask );
       }
 CODE
@@ -412,7 +447,7 @@ sub rotate_shift ($$) {
 	print << "CODE";
       {
 	libspectrum_byte bytetemp = readbyte(HL);
-	contend_read( HL, 1 );
+	contend_read_no_mreq( HL, 1 );
 	$opcode(bytetemp);
 	writebyte(HL,bytetemp);
       }
@@ -421,7 +456,7 @@ CODE
 	print << "CODE";
       {
 	libspectrum_byte bytetemp = readbyte(tempaddr);
-	contend_read( tempaddr, 1 );
+	contend_read_no_mreq( tempaddr, 1 );
 	$opcode(bytetemp);
 	writebyte(tempaddr,bytetemp);
       }
@@ -447,15 +482,15 @@ sub opcode_BIT (@) {
 	print << "BIT";
       {
 	libspectrum_byte bytetemp = readbyte( tempaddr );
-	contend_read( tempaddr, 1 );
-	BIT( $bit, bytetemp );
+	contend_read_no_mreq( tempaddr, 1 );
+	BIT_I( $bit, bytetemp, tempaddr );
       }
 BIT
     } else {
 	print << "BIT";
       {
 	libspectrum_byte bytetemp = readbyte( HL );
-	contend_read( HL, 1 );
+	contend_read_no_mreq( HL, 1 );
 	BIT( $bit, bytetemp );
       }
 BIT
@@ -493,16 +528,15 @@ sub opcode_DAA (@) {
     print << "DAA";
       {
 	libspectrum_byte add = 0, carry = ( F & FLAG_C );
-	if( ( F & FLAG_H ) || ( (A & 0x0f)>9 ) ) add=6;
-	if( carry || (A > 0x9f ) ) add|=0x60;
-	if( A > 0x99 ) carry=1;
-	if ( F & FLAG_N ) {
+	if( ( F & FLAG_H ) || ( ( A & 0x0f ) > 9 ) ) add = 6;
+	if( carry || ( A > 0x99 ) ) add |= 0x60;
+	if( A > 0x99 ) carry = FLAG_C;
+	if( F & FLAG_N ) {
 	  SUB(add);
 	} else {
-	  if( (A>0x90) && ( (A & 0x0f)>9) ) add|=0x60;
 	  ADD(add);
 	}
-	F = ( F & ~( FLAG_C | FLAG_P) ) | carry | parity_table[A];
+	F = ( F & ~( FLAG_C | FLAG_P ) ) | carry | parity_table[A];
       }
 DAA
 }
@@ -513,7 +547,7 @@ sub opcode_DI (@) { print "      IFF1=IFF2=0;\n"; }
 
 sub opcode_DJNZ (@) {
     print << "DJNZ";
-      tstates++;
+      contend_read_no_mreq( IR, 1 );
       B--;
       if(B) {
 	JR();
@@ -530,7 +564,7 @@ sub opcode_EI (@) {
 	 accepted after the next instruction */
       IFF1 = IFF2 = 1;
       z80.interrupts_enabled_at = tstates;
-      event_add( tstates + 1, EVENT_TYPE_INTERRUPT );
+      event_add( tstates + 1, z80_interrupt_event );
 EI
 }
 
@@ -565,10 +599,10 @@ EX
       {
 	libspectrum_byte bytetempl, bytetemph;
 	bytetempl = readbyte( SP );
-	bytetemph = readbyte( SP + 1 ); tstates++;
-	writebyte( SP,     $low  );
+	bytetemph = readbyte( SP + 1 ); contend_read_no_mreq( SP + 1, 1 );
 	writebyte( SP + 1, $high );
-	contend_write( SP + 1, 1 ); contend_write( SP + 1, 1 );
+	writebyte( SP,     $low  );
+	contend_write_no_mreq( SP, 1 ); contend_write_no_mreq( SP, 1 );
 	$low=bytetempl; $high=bytetemph;
       }
 EX
@@ -610,7 +644,6 @@ sub opcode_IN (@) {
       { 
 	libspectrum_word intemp;
 	intemp = readbyte( PC++ ) + ( A << 8 );
-	ula_contend_port( intemp );
         A=readport( intemp );
       }
 IN
@@ -685,25 +718,27 @@ sub opcode_LD (@) {
 
 	if( length $src == 1 or $src =~ /^REGISTER[HL]$/ ) {
 
+	    if( $dest eq 'R' or $src eq 'R' or $dest eq 'I' or $src eq 'I') {
+		print "      contend_read_no_mreq( IR, 1 );\n"
+	    }
+
 	    if( $dest eq 'R' and $src eq 'A' ) {
 		print << "LD";
-      tstates += 1;
       /* Keep the RZX instruction counter right */
       rzx_instructions_offset += ( R - A );
       R=R7=A;
 LD
             } elsif( $dest eq 'A' and $src eq 'R' ) {
-		print << "LD";
-      tstates += 1;
-      A=(R&0x7f) | (R7&0x80);
-      F = ( F & FLAG_C ) | sz53_table[A] | ( IFF2 ? FLAG_V : 0 );
-LD
+		print "      A=(R&0x7f) | (R7&0x80);\n";
 	    } else {
-		print "      tstates += 1;\n" if $src eq 'I' or $dest eq 'I';
 		print "      $dest=$src;\n" if $dest ne $src;
-		if( $dest eq 'A' and $src eq 'I' ) {
-		    print "      F = ( F & FLAG_C ) | sz53_table[A] | ( IFF2 ? FLAG_V : 0 );\n";
-		}
+	    }
+            if( $dest eq 'A' and ( $src eq 'I' or $src eq 'R' ) ) {
+		print << "LD";
+      F = ( F & FLAG_C ) | sz53_table[A] | ( IFF2 ? FLAG_V : 0 );
+      z80.iff2_read = 1;
+      event_add( tstates, z80_nmos_iff2_event );
+LD
 	    }
 	} elsif( $src eq 'nn' ) {
 	    print "      $dest = readbyte( PC++ );\n";
@@ -726,8 +761,9 @@ LD
       {
 	libspectrum_byte offset;
 	offset = readbyte( PC );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); contend_read( PC, 1 );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); PC++;
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 );
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 );
+	contend_read_no_mreq( PC, 1 ); PC++;
 	$dest = readbyte( REGISTER + (libspectrum_signed_byte)offset );
       }
 LD
@@ -750,7 +786,11 @@ LD
       $high=readbyte(PC++);
 LD
         } elsif( $src eq 'HL' or $src eq 'REGISTER' ) {
-	    print "      tstates += 2;\n      SP=$src;\n";
+	    print << "LD";
+      contend_read_no_mreq( IR, 1 );
+      contend_read_no_mreq( IR, 1 );
+      SP = $src;
+LD
         } elsif( $src eq '(nnnn)' ) {
 	    print "      LD16_RRNN($low,$high);\n";
 	}
@@ -798,8 +838,9 @@ LD
       {
 	libspectrum_byte offset;
 	offset = readbyte( PC );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); contend_read( PC, 1 );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); PC++;
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 );
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 );
+	contend_read_no_mreq( PC, 1 ); PC++;
 	writebyte( REGISTER + (libspectrum_signed_byte)offset, $src );
       }
 LD
@@ -809,7 +850,7 @@ LD
 	libspectrum_byte offset, value;
 	offset = readbyte( PC++ );
 	value = readbyte( PC );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); PC++;
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 ); PC++;
 	writebyte( REGISTER + (libspectrum_signed_byte)offset, value );
       }
 LD
@@ -853,13 +894,15 @@ sub opcode_OUT (@) {
       { 
 	libspectrum_word outtemp;
 	outtemp = readbyte( PC++ ) + ( A << 8 );
-	Z80_OUT( outtemp , A );
+	writeport( outtemp, A );
       }
 OUT
     } elsif( $port eq '(C)' and length $register == 1 ) {
-	print << "OUT";
-      Z80_OUT( BC, $register );
-OUT
+	if ( $register eq '0' ) {
+	    print "      writeport( BC, IS_CMOS ? 0xff : 0 );\n";
+	} else {
+	    print "      writeport( BC, $register );\n";
+	}
     }
 }
 
@@ -874,7 +917,7 @@ sub opcode_PUSH (@) {
 
     my( $regpair ) = @_;
 
-    print "      tstates++;\n";
+    print "      contend_read_no_mreq( IR, 1 );\n";
     push_pop( 'PUSH', $regpair );
 }
 
@@ -887,7 +930,7 @@ sub opcode_RET (@) {
     if( not defined $condition ) {
 	print "      RET();\n";
     } else {
-	print "      tstates++;\n";
+	print "      contend_read_no_mreq( IR, 1 );\n";
 	
 	if( $condition eq 'NZ' ) {
 	    print << "RET";
@@ -910,6 +953,7 @@ sub opcode_RETN (@) {
     print << "RETN";
       IFF1=IFF2;
       RET();
+      z80_retn();
 RETN
 }
 
@@ -940,8 +984,8 @@ sub opcode_RLD (@) {
     print << "RLD";
       {
 	libspectrum_byte bytetemp = readbyte( HL );
-	contend_read( HL, 1 ); contend_read( HL, 1 ); contend_read( HL, 1 );
-	contend_read( HL, 1 );
+	contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
+	contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
 	writebyte(HL, (bytetemp << 4 ) | ( A & 0x0f ) );
 	A = ( A & 0xf0 ) | ( bytetemp >> 4 );
 	F = ( F & FLAG_C ) | sz53p_table[A];
@@ -976,8 +1020,8 @@ sub opcode_RRD (@) {
     print << "RRD";
       {
 	libspectrum_byte bytetemp = readbyte( HL );
-	contend_read( HL, 1 ); contend_read( HL, 1 ); contend_read( HL, 1 );
-	contend_read( HL, 1 );
+	contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
+	contend_read_no_mreq( HL, 1 ); contend_read_no_mreq( HL, 1 );
 	writebyte(HL,  ( A << 4 ) | ( bytetemp >> 4 ) );
 	A = ( A & 0xf0 ) | ( bytetemp & 0x0f );
 	F = ( F & FLAG_C ) | sz53p_table[A];
@@ -989,7 +1033,7 @@ sub opcode_RST (@) {
 
     my( $value ) = @_;
 
-    printf "      tstates++;\n      RST(0x%02x);\n", hex $value;
+    printf "      contend_read_no_mreq( IR, 1 );\n      RST(0x%02x);\n", hex $value;
 }
 
 sub opcode_SBC (@) { arithmetic_logical( 'SBC', $_[0], $_[1] ); }
@@ -1036,7 +1080,7 @@ sub opcode_shift (@) {
 	    REGISTER + (libspectrum_signed_byte)readbyte_internal( PC );
 	PC++; contend_read( PC, 3 );
 	opcode3 = readbyte_internal( PC );
-	contend_read( PC, 1 ); contend_read( PC, 1 ); PC++;
+	contend_read_no_mreq( PC, 1 ); contend_read_no_mreq( PC, 1 ); PC++;
 #ifdef HAVE_ENOUGH_MEMORY
 	switch(opcode3) {
 #include "z80_ddfdcb.c"
@@ -1075,7 +1119,7 @@ shift
         print << "shift"
 	}
 #else			/* #ifdef HAVE_ENOUGH_MEMORY */
-	z80_${lc_opcode}xx(opcode2);
+	if( z80_${lc_opcode}xx(opcode2) ) goto end_opcode;
 #endif			/* #ifdef HAVE_ENOUGH_MEMORY */
       }
 shift
@@ -1086,17 +1130,17 @@ shift
 
 my %description = (
 
-    'opcodes_cb.dat'     => 'opcodes_cb.c: Z80 CBxx opcodes',
-    'opcodes_ddfd.dat'   => 'opcodes_ddfd.c Z80 {DD,FD}xx opcodes',
-    'opcodes_ddfdcb.dat' => 'opcodes_ddfdcb.c Z80 {DD,FD}CBxx opcodes',
-    'opcodes_ed.dat'     => 'opcodes_ed.c: Z80 CBxx opcodes',
+    'opcodes_cb.dat'     => 'z80_cb.c: Z80 CBxx opcodes',
+    'opcodes_ddfd.dat'   => 'z80_ddfd.c Z80 {DD,FD}xx opcodes',
+    'opcodes_ddfdcb.dat' => 'z80_ddfdcb.c Z80 {DD,FD}CBxx opcodes',
+    'opcodes_ed.dat'     => 'z80_ed.c: Z80 CBxx opcodes',
     'opcodes_base.dat'   => 'opcodes_base.c: unshifted Z80 opcodes',
 
 );
 
 # Main program
 
-my $data_file = $ARGV[0];
+( my $data_file = $ARGV[0] ) =~ s!.*/!!;
 
 print Fuse::GPL( $description{ $data_file }, '1999-2003 Philip Kendall' );
 
@@ -1150,7 +1194,7 @@ while(<>) {
 
 	    print << "CODE";
       $register = readbyte(tempaddr) $operator $hexmask;
-      contend_read( tempaddr, 1 );
+      contend_read_no_mreq( tempaddr, 1 );
       writebyte(tempaddr, $register);
       break;
 CODE
@@ -1158,7 +1202,7 @@ CODE
 
 	    print << "CODE";
       $register=readbyte(tempaddr);
-      contend_read( tempaddr, 1 );
+      contend_read_no_mreq( tempaddr, 1 );
       $opcode($register);
       writebyte(tempaddr, $register);
       break;
@@ -1183,9 +1227,14 @@ if( $data_file eq 'opcodes_ddfd.dat' ) {
     print << "CODE";
     default:		/* Instruction did not involve H or L, so backtrack
 			   one instruction and parse again */
-      PC--;		/* FIXME: will be contended again */
-      R--;		/* Decrement the R register as well */
-      break;
+      PC--;
+      R--;
+      opcode = opcode2;
+#ifdef HAVE_ENOUGH_MEMORY
+      goto end_opcode;
+#else			/* #ifdef HAVE_ENOUGH_MEMORY */
+      return 1;
+#endif			/* #ifdef HAVE_ENOUGH_MEMORY */
 CODE
 
 } elsif( $data_file eq 'opcodes_ed.dat' ) {
