@@ -1,7 +1,8 @@
 /* z80_ops.c: Process the next opcode
    Copyright (c) 1999-2005 Philip Kendall, Witold Filipczyk
-
-   $Id: z80_ops.c 3681 2008-06-16 09:40:29Z pak21 $
+   Copyright (c) 2015 Stuart Brady
+   Copyright (c) 2015 Gergely Szasz
+   Copyright (c) 2015 Sergio Baldoví
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,20 +29,26 @@
 #include <stdio.h>
 
 #include "debugger/debugger.h"
-#include "disk/beta.h"
-#include "disk/plusd.h"
-#include "divide.h"
 #include "event.h"
 #include "machine.h"
 #include "memory.h"
 #include "periph.h"
+#include "peripherals/disk/beta.h"
+#include "peripherals/disk/didaktik.h"
+#include "peripherals/disk/disciple.h"
+#include "peripherals/disk/opus.h"
+#include "peripherals/disk/plusd.h"
+#include "peripherals/ide/divide.h"
+#include "peripherals/if1.h"
+#include "peripherals/spectranet.h"
+#include "peripherals/ula.h"
+#include "peripherals/usource.h"
 #include "profile.h"
 #include "rzx.h"
-#include "if1.h"
 #include "settings.h"
 #include "slt.h"
+#include "svg.h"
 #include "tape.h"
-#include "ula.h"
 #include "z80.h"
 
 #include "z80_macros.h"
@@ -159,13 +166,16 @@ z80_do_opcodes( void )
 
     CHECK( beta, beta_available )
 
+#define NOT_128_TYPE_OR_IS_48_TYPE ( !( machine_current->capabilities & \
+            LIBSPECTRUM_MACHINE_CAPABILITY_128_MEMORY ) || \
+            machine_current->ram.current_rom )
+
     if( beta_active ) {
-      if( machine_current->ram.current_rom &&
-	  PC >= 16384 ) {
+      if( NOT_128_TYPE_OR_IS_48_TYPE && PC >= 16384 ) {
 	beta_unpage();
       }
-    } else if( ( PC & 0xff00 ) == 0x3d00 &&
-	       machine_current->ram.current_rom ) {
+    } else if( ( PC & beta_pc_mask ) == beta_pc_value &&
+               NOT_128_TYPE_OR_IS_48_TYPE ) {
       beta_page();
     }
 
@@ -179,6 +189,32 @@ z80_do_opcodes( void )
 
     END_CHECK
 
+    CHECK( didaktik80, didaktik80_available )
+
+    if( PC == 0x0000 || PC == 0x0008 ) {
+      didaktik80_page();
+    } else if( PC == 0x1700 ) {
+      didaktik80_unpage();
+    }
+
+    END_CHECK
+
+    CHECK( disciple, disciple_available )
+
+    if( PC == 0x0001 || PC == 0x0008 || PC == 0x0066 || PC == 0x028e ) {
+      disciple_page();
+    }
+
+    END_CHECK
+
+    CHECK( usource, usource_available )
+
+    if( PC == 0x2bae ) {
+      usource_toggle();
+    }
+
+    END_CHECK
+
     CHECK( if1p, if1_available )
 
     if( PC == 0x0008 || PC == 0x1708 ) {
@@ -195,6 +231,17 @@ z80_do_opcodes( void )
     
     END_CHECK
 
+    CHECK( spectranet_page, spectranet_available && !settings_current.spectranet_disable )
+
+    if( PC == 0x0008 || ((PC & 0xfff8) == 0x3ff8) )
+      spectranet_page( 0 );
+
+    if( PC == spectranet_programmable_trap &&
+      spectranet_programmable_trap_active )
+      event_add( 0, z80_nmi_event );
+
+    END_CHECK
+
   opcode_delay:
 
     contend_read( PC, 4 );
@@ -202,7 +249,11 @@ z80_do_opcodes( void )
     /* Check to see if M1 cycles happen on even tstates */
     CHECK( evenm1, even_m1 )
 
-    if( tstates & 1 ) tstates++;
+    if( tstates & 1 ) {
+      if( ++tstates == event_next_event ) {
+	break;
+      }
+    }
 
     END_CHECK
 
@@ -230,10 +281,52 @@ z80_do_opcodes( void )
     
     END_CHECK
 
+    CHECK( opus, opus_available )
+
+    if( opus_active ) {
+      if( PC == 0x1748 ) {
+        opus_unpage();
+      }
+    } else if( PC == 0x0008 || PC == 0x0048 || PC == 0x1708 ) {
+      opus_page();
+    }
+
+    END_CHECK
+
+    CHECK( spectranet_unpage, spectranet_available )
+
+    if( PC == 0x007c )
+      spectranet_unpage();
+
+    END_CHECK
+
+    CHECK( z80_iff2_read, z80.iff2_read )
+
+    z80.iff2_read = 0;
+    /* Execute *one* instruction before reevaluating the checks */
+    event_add( tstates, z80_nmos_iff2_event );
+
+    END_CHECK
+
+    CHECK( didaktik80snap, didaktik80_snap )
+
+    if( PC == 0x0066 && !didaktik80_active ) {
+      opcode = 0xc7;	/* RST 00 */
+      didaktik80_snap = 0; /* FIXME: this should be a time-based reset */
+    }
+
+    END_CHECK
+
+    CHECK( svg_capture, svg_capture_active )
+
+    svg_capture();
+
+    END_CHECK
+
   end_opcode:
     PC++; R++;
     switch(opcode) {
-#include "opcodes_base.c"
+#include "z80/opcodes_base.c"
     }
 
   }
@@ -246,7 +339,7 @@ static int
 z80_cbxx( libspectrum_byte opcode2 )
 {
   switch(opcode2) {
-#include "z80_cb.c"
+#include "z80/z80_cb.c"
   }
   return 0;
 }
@@ -258,7 +351,7 @@ z80_ddxx( libspectrum_byte opcode2 )
 #define REGISTER  IX
 #define REGISTERL IXL
 #define REGISTERH IXH
-#include "z80_ddfd.c"
+#include "z80/z80_ddfd.c"
 #undef REGISTERH
 #undef REGISTERL
 #undef REGISTER
@@ -270,7 +363,7 @@ static int
 z80_edxx( libspectrum_byte opcode2 )
 {
   switch(opcode2) {
-#include "z80_ed.c"
+#include "z80/z80_ed.c"
   }
   return 0;
 }
@@ -282,7 +375,7 @@ z80_fdxx( libspectrum_byte opcode2 )
 #define REGISTER  IY
 #define REGISTERL IYL
 #define REGISTERH IYH
-#include "z80_ddfd.c"
+#include "z80/z80_ddfd.c"
 #undef REGISTERH
 #undef REGISTERL
 #undef REGISTER
@@ -294,7 +387,7 @@ static void
 z80_ddfdcbxx( libspectrum_byte opcode3, libspectrum_word tempaddr )
 {
   switch(opcode3) {
-#include "z80_ddfdcb.c"
+#include "z80/z80_ddfdcb.c"
   }
 }
 
